#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""


class SubmitContext:
    """[summary]"""

    def submit_model(self):
        """[summary]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def wait_for_completion(self):
        """[summary]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def download_files(self):
        """[summary]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from typing import Tuple

import click
import click_spinner
from azureml.core.workspace import Workspace

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import discover_model
from energinetml.core.model import Model
from energinetml.settings import DEFAULT_VM_CPU, DEFAULT_VM_GPU


@click.command()
@discover_model()
@click.option(
    "--min-nodes",
    "min_nodes",
    required=False,
    default=None,
    type=int,
    help="Min. number of compute nodes, if creating a new cluster",
)
@click.option(
    "--max-nodes",
    "max_nodes",
    required=False,
    default=None,
    type=int,
    help="Max. number of compute nodes, if creating a new cluster",
)
@click.option(
    "--cluster-name",
    "cluster_name",
    required=False,
    default=None,
    type=str,
    help="Name of compute cluster, if creating a new cluster",
)
@click.option(
    "--specify-vm/--default",
    "specify_vm",
    default=None,
    help=(
        "Whether to specify exact VM size, or use default for "
        "either CPU or GPU computation, if creating a new cluster"
    ),
)
@click.option(
    "--vm-size",
    "vm_size",
    required=False,
    default=None,
    type=str,
    help="VM Size to use, if using the --specify-vm parameter",
)
@click.option(
    "--cpu/--gpu",
    "cpu",
    default=None,
    help=(
        "Whether to use default CPU- og GPU VM-size, if "
        "using the --default parameter"
    ),
)
def create(
    model: Model,
    min_nodes: int,
    max_nodes: int,
    cluster_name: str,
    specify_vm: bool,
    vm_size: str,
    cpu: str,
) -> None:
    """Create a new compute cluster and attach it to the model."""

    project_meta = model.project.as_dict()
    workspace = backend.get_workspace(project_meta)

    vm_size, default_cluster_name = _get_vm_size_and_cluster_name(
        workspace=workspace, specify_vm=specify_vm, vm_size=vm_size, cpu=cpu
    )

    if min_nodes is None:
        min_nodes = click.prompt(
            text="Please enter minimum nodes available", default=0, type=click.INT
        )

    if max_nodes is None:
        max_nodes = click.prompt(
            text="Please enter maximum nodes available", default=1, type=click.INT
        )

    if cluster_name is None:
        cluster_name = click.prompt(
            text="Please enter a name for the compute cluster",
            default=default_cluster_name,
            type=click.STRING,
        )

    cluster_name = _normalize_cluster_name(cluster_name)
    existing_clusters = backend.get_compute_clusters(workspace)
    existing_cluster_names = [c.name for c in existing_clusters]

    while cluster_name in existing_cluster_names:
        click.echo(f"Cluster already exists: {cluster_name}")
        cluster_name = _normalize_cluster_name(
            click.prompt(
                text="Please enter a name for the compute cluster", type=click.STRING
            )
        )

    click.echo(f'Creating compute cluster "{cluster_name}" using VM Size: {vm_size}')

    with click_spinner.spinner():
        backend.create_compute_cluster(
            workspace=workspace,
            name=cluster_name,
            vm_size=vm_size,
            min_nodes=min_nodes,
            max_nodes=max_nodes,
            vnet_resource_group_name=model.project.resource_group,
            vnet_name=model.project.vnet_name,
            subnet_name=model.project.subnet_name,
        )

    _update_model_properties(model=model, cluster=cluster_name, vm_size=vm_size)


def _update_model_properties(model: Model, cluster: str, vm_size: str) -> None:
    """[summary]

    Args:
        model (Model): [description]
        cluster (str): [description]
        vm_size (str): [description]
    """
    model.compute_target = cluster
    model.vm_size = vm_size
    model.save()


# -- Helper functions --------------------------------------------------------


def _get_vm_size_and_cluster_name(
    workspace: Workspace,
    specify_vm: bool,
    vm_size: str,
    cpu: str,
) -> Tuple[str, str]:
    """[summary]

    Args:
        workspace (Workspace): [description]
        specify_vm (bool): [description]
        vm_size (str): [description]
        cpu (str): [description]

    Returns:
        Tuple[str, str]: [description]
    """
    available_vm_sizes = backend.get_available_vm_sizes(workspace)
    available_vm_size_mapped = {vm["name"]: vm for vm in available_vm_sizes}

    if specify_vm is None:
        click.echo(
            "You can either specific an exact VM Size, or use a default "
            "VM Size for either CPU or GPU computation."
        )
        specify_vm = (
            click.prompt(
                text="How would you like to specify VM Size, or use a default?",
                type=click.Choice(["vmsize", "default"]),
            )
            == "vmsize"
        )

    if specify_vm:
        vm_size, cluster_name = _get_specific_vm_size(workspace, vm_size)
    else:
        vm_size, cluster_name = _get_default_vm_size(cpu)

    while vm_size not in available_vm_size_mapped:
        click.echo("VM Size unavailable: %s" % vm_size)
        vm_size = click.prompt(
            text="Please enter a VM size", type=click.Choice(available_vm_size_mapped)
        )

    return vm_size, cluster_name


def _get_specific_vm_size(workspace: Workspace, vm_size: str) -> Tuple[str, str]:
    """[summary]

    Args:
        workspace (Workspace): [description]
        vm_size (str): [description]

    Returns:
        Tuple[str, str]: [description]
    """
    available_vm_sizes = backend.get_available_vm_sizes(workspace)
    available_vm_size_mapped = {vm["name"]: vm for vm in available_vm_sizes}

    if vm_size is None:
        vm_size = click.prompt(
            text="Please enter a VM size", type=click.Choice(available_vm_size_mapped)
        )

    return vm_size, _normalize_cluster_name(vm_size)


def _get_default_vm_size(cpu: bool) -> Tuple[str, str]:
    """[summary]

    Args:
        cpu (bool): [description]

    Returns:
        Tuple[str, str]: [description]
    """
    if cpu is None:
        cpu = (
            click.prompt(
                text="Which kind of computing would you like to use?",
                type=click.Choice(["cpu", "gpu"]),
            )
            == "cpu"
        )

    if cpu:
        return DEFAULT_VM_CPU, "CPU-Cluster"
    else:
        return DEFAULT_VM_GPU, "GPU-Cluster"


def _normalize_cluster_name(cluster_name: str) -> str:
    """[summary]

    Args:
        cluster_name (str): [description]

    Returns:
        str: [description]
    """
    return cluster_name.replace("_", "-")



from unittest.mock import ANY, Mock, PropertyMock, patch

import pytest
from click.testing import CliRunner

from energinetml.cli.cluster.create import (
    DEFAULT_VM_CPU,
    DEFAULT_VM_GPU,
    _get_default_vm_size,
    _get_specific_vm_size,
    _get_vm_size_and_cluster_name,
    _update_model_properties,
    create,
)
from tests.constants import CLUSTER_NAME, RESOURCE_GROUP, VM_SIZE, WORKSPACE_NAME

# -- create() Tests ----------------------------------------------------------


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._get_vm_size_and_cluster_name", name="asd")
@patch("energinetml.cli.cluster.create._normalize_cluster_name")
@patch("energinetml.cli.cluster.create._update_model_properties")
def test__create_cluster__should_create_new_cluster(
    _update_model_properties_mock,
    _normalize_cluster_name_mock,
    _get_vm_size_and_cluster_name_mock,
    backend_mock,
    model_with_project,
):
    """
    :param Mock _update_model_properties_mock:
    :param Mock _normalize_cluster_name_mock:
    :param Mock _get_vm_size_and_cluster_name_mock:
    :param Mock backend_mock:
    :param Model model_with_project:
    """
    backend_mock.get_workspace.return_value = WORKSPACE_NAME
    _get_vm_size_and_cluster_name_mock.return_value = VM_SIZE, "existing-cluster"
    _normalize_cluster_name_mock.return_value = "actual-cluster-name"

    # Workaround: Can not mock property "name" of Mock object
    cluster = Mock()
    type(cluster).name = PropertyMock(return_value="existing-cluster")

    backend_mock.get_compute_clusters.return_value = [cluster]

    # Act
    result = CliRunner().invoke(
        cli=create,
        args=["--path", model_with_project.path],
        # input='%s\n' % CLUSTER_NAME,
        input="\n".join(
            ["123", "456", CLUSTER_NAME]  # min_nodes  # max_nodes  # cluster_name
        ),
    )

    # Assert
    assert result.exit_code == 0

    backend_mock.create_compute_cluster.assert_called_once_with(
        workspace=WORKSPACE_NAME,
        name="actual-cluster-name",
        vm_size=VM_SIZE,
        min_nodes=123,
        max_nodes=456,
        vnet_resource_group_name=RESOURCE_GROUP,
        vnet_name=ANY,
        subnet_name=ANY,
    )

    _update_model_properties_mock.assert_called_once_with(
        model=ANY,  # Model is not the same instance as defined above
        cluster="actual-cluster-name",
        vm_size=VM_SIZE,
    )


# -- _update_model_properties() Tests ----------------------------------------


def test__update_model_properties__should_set_properties_and_save():
    # Arrange
    model = Mock()

    # Act
    _update_model_properties(model=model, cluster="cluster", vm_size="vm-size")

    # Assert
    assert model.compute_target == "cluster"
    assert model.vm_size == "vm-size"
    model.save.assert_called_once()


# -- _update_model_properties() Tests ----------------------------------------


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._get_specific_vm_size")
@patch("energinetml.cli.cluster.create._get_default_vm_size")
@patch("energinetml.cli.cluster.create.click.prompt")
def test__get_vm_size_and_cluster_name__should_prompt_to_specify_vm(
    prompt_mock, _get_default_vm_size_mock, _get_specific_vm_size_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    # Arrange
    _get_default_vm_size_mock.return_value = "VM_1", "cluster-name-1"
    _get_specific_vm_size_mock.return_value = "VM_2", "cluster-name-2"
    backend_mock.get_available_vm_sizes.return_value = [
        {"name": "VM_1"},
        {"name": "VM_2"},
    ]

    # Act
    vm_size, cluster_name = _get_vm_size_and_cluster_name(
        workspace=WORKSPACE_NAME, specify_vm=None, vm_size="vm-size", cpu=False
    )

    # Assert
    prompt_mock.assert_called_once_with(
        text="How would you like to specify VM Size, or use a default?", type=ANY
    )


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._get_specific_vm_size")
@patch("energinetml.cli.cluster.create._get_default_vm_size")
@patch("energinetml.cli.cluster.create.click.prompt")
def test__get_vm_size_and_cluster_name__specify_vm__should_return_vm_size_and_cluster_name(  # noqa: E501
    prompt_mock, _get_default_vm_size_mock, _get_specific_vm_size_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    # Arrange
    _get_default_vm_size_mock.return_value = "VM_1", "cluster-name-1"
    _get_specific_vm_size_mock.return_value = "VM_2", "cluster-name-2"
    backend_mock.get_available_vm_sizes.return_value = [
        {"name": "VM_1"},
        {"name": "VM_2"},
    ]

    # Act
    vm_size, cluster_name = _get_vm_size_and_cluster_name(
        workspace=WORKSPACE_NAME, specify_vm=True, vm_size="vm-size", cpu=False
    )

    # Assert
    prompt_mock.assert_not_called()
    _get_default_vm_size_mock.assert_not_called()
    _get_specific_vm_size_mock.assert_called_once_with(WORKSPACE_NAME, "vm-size")
    assert vm_size == "VM_2"
    assert cluster_name == "cluster-name-2"


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._get_specific_vm_size")
@patch("energinetml.cli.cluster.create._get_default_vm_size")
@patch("energinetml.cli.cluster.create.click.prompt")
def test__get_vm_size_and_cluster_name__use_default_vm__should_return_vm_size_and_cluster_name(  # noqa: E501
    prompt_mock, _get_default_vm_size_mock, _get_specific_vm_size_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    # Arrange
    _get_default_vm_size_mock.return_value = "VM_1", "cluster-name-1"
    _get_specific_vm_size_mock.return_value = "VM_2", "cluster-name-2"
    backend_mock.get_available_vm_sizes.return_value = [
        {"name": "VM_1"},
        {"name": "VM_2"},
    ]

    # Act
    vm_size, cluster_name = _get_vm_size_and_cluster_name(
        workspace=WORKSPACE_NAME, specify_vm=False, vm_size="vm-size", cpu=False
    )

    # Assert
    prompt_mock.assert_not_called()
    _get_specific_vm_size_mock._get_default_vm_size_mock(False)
    _get_specific_vm_size_mock.assert_not_called()
    assert vm_size == "VM_1"
    assert cluster_name == "cluster-name-1"


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._get_specific_vm_size")
@patch("energinetml.cli.cluster.create._get_default_vm_size")
@patch("energinetml.cli.cluster.create.click.prompt")
def test__get_vm_size_and_cluster_name__vm_unavailable__should_prompt_for_vm_size(  # noqa: E501
    prompt_mock, _get_default_vm_size_mock, _get_specific_vm_size_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock backend_mock:
    """
    # Arrange
    _get_default_vm_size_mock.return_value = "VM-NAME-DOES-NOT-EXIST", "cluster-name-1"
    _get_specific_vm_size_mock.return_value = "VM_2", "cluster-name-2"
    backend_mock.get_available_vm_sizes.return_value = [
        {"name": "VM_1"},
        {"name": "VM_2"},
    ]

    prompt_mock.return_value = "VM_1"

    # Act
    vm_size, cluster_name = _get_vm_size_and_cluster_name(
        workspace=WORKSPACE_NAME, specify_vm=False, vm_size="vm-size", cpu=False
    )

    # Assert
    prompt_mock.assert_called_once_with(text="Please enter a VM size", type=ANY)
    _get_specific_vm_size_mock._get_default_vm_size_mock(False)
    _get_specific_vm_size_mock.assert_not_called()
    assert vm_size == "VM_1"
    assert cluster_name == "cluster-name-1"


# -- _get_specific_vm_size() Tests -------------------------------------------


@patch("energinetml.cli.cluster.create.backend")
@patch("energinetml.cli.cluster.create._normalize_cluster_name")
@patch("energinetml.cli.cluster.create.click.prompt")
def test__get_specific_vm_size__should_prompt_for_vm_size_and_return_normalized_name(
    prompt_mock, _normalize_cluster_name_mock, backend_mock
):
    """
    :param Mock prompt_mock:
    :param Mock _normalize_cluster_name_mock:
    :param Mock backend_mock:
    """
    # Arrange
    _normalize_cluster_name_mock.return_value = "normalized-cluster-name"
    backend_mock.get_available_vm_sizes.return_value = [
        {"name": "VM_1"},
        {"name": "VM_2"},
    ]

    prompt_mock.return_value = "VM_1"

    # Act
    vm_size, cluster_name = _get_specific_vm_size(
        workspace=WORKSPACE_NAME, vm_size=None
    )

    # Assert
    prompt_mock.assert_called_once_with(text="Please enter a VM size", type=ANY)
    assert vm_size == "VM_1"
    assert cluster_name == "normalized-cluster-name"


# -- _get_default_vm_size() Tests --------------------------------------------


@patch("energinetml.cli.cluster.create.click.prompt")
@pytest.mark.parametrize(
    "prompt_value, vm_size, cluster_name",
    (("cpu", DEFAULT_VM_CPU, "CPU-Cluster"), ("gpu", DEFAULT_VM_GPU, "GPU-Cluster")),
)
def test__get_default_vm_size__should_prompt_for_vm_type_and_return_vm_size_and_name(
    prompt_mock, prompt_value, vm_size, cluster_name
):
    """
    :param Mock prompt_mock:
    """
    # Arrange
    prompt_mock.return_value = prompt_value

    # Act
    actual_vm_size, actual_cluster_name = _get_default_vm_size(cpu=None)

    # Assert
    assert actual_vm_size == vm_size
    assert actual_cluster_name == cluster_name



"""A module for the feature validator"""

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import numpy as np
import pandas as pd  # type: ignore
import pydantic


def get_uuid() -> str:
    """A function which returns a uuid4 as a string."""
    return str(uuid4())


class ValidatorStatus(str, enum.Enum):
    """Current possibly status of the validator."""

    VALID = enum.auto()
    WARNING = enum.auto()
    NOT_VALID = enum.auto()
    NOT_USED = enum.auto()


class ValidatorResponse(pydantic.BaseModel):
    """A container to hold validator information."""

    state: Union[ValidatorStatus, str]
    message: Optional[str]


class FeatureType(enum.Enum):
    """Currently supporten types. This enum is used when doing the dynamic pydantic
    Model creation inside pydantic.create_model()."""

    INT = int
    FLOAT = float
    STRING = str
    CATEGORY = str
    DATETIME = datetime


class FeatureMetaDataMissing(Exception):
    """The class raised if meta_data is not valid."""

    def __init__(self, name: str, message: str) -> None:
        self.name = name
        self.message = message
        super().__init__(message)


class FeatureValueError(ValueError):
    """The class is raised if feature validator gets a not valid input value."""

    def __init__(self, state: ValidatorStatus, message: str) -> None:
        self.state = state
        super().__init__(message)


@dataclass
class Validator:
    """The validator class contains one or several feature validator(s)."""

    dataset_id: str
    feature_models: List[
        Union[
            "IntegerFeature",
            "FloatFeature",
            "StringFeature",
            "CategoricalFeature",
            "DatetimeFeature",
        ]
    ]

    validator_uuid: str = field(default_factory=get_uuid)

    def get_features(self) -> Dict[str, Tuple["FeatureType", ...]]:
        """The get_feature() method can be unwrapped and used as input arguments
        in the dynamic model with the create_model() method.

        Returns:
            Dict[str, Tuple[FeatureType, ...]]: A dictionary with key as feature name
            and value as tuple with feature type and dots (...) for feature value.
        """
        return {
            feature.name: (feature.feature_type.value, ...)
            for feature in self.feature_models
        }

    def get_validators(self) -> Dict[str, pydantic.validator]:
        """The get_validators() method can be used as a validator when instantiating
        the dynamic model with the create_model() method.

        Returns:
            Dict[str, validator]: A dictionary with key as validatorname name
            and value as the validator for the given feature.
        """
        prefix: str = "_check_"

        return {
            f"{prefix}{feature.name}": feature.get_validator()
            for feature in self.feature_models
        }

    def get_pydantic_model(
        self, model_name: str = "PredictFeatures", use_validation: bool = True
    ) -> pydantic.BaseModel:
        """This method wraps the get_feature method and the get _valdiators method into
        a pydantic.BaseModel object.

        Args:
            model_name (str, optional): Name of the pydantic.BaseModel.
            Defaults to "PredictFeatures".
            use_validation (bool, optional): Boolean argument for including a validator.
            Defaults to True.

        Returns:
            pydantic.BaseModel: A pydantic.BaseModel which corresponds to the
            input features for a given model.
        """

        return pydantic.create_model(
            __model_name=model_name,
            **self.get_features(),
            __base__=pydantic.BaseModel,
            __validators__=self.get_validators() if use_validation else None,
        )


def create_feature_models(
    feature_names: List[str],
    feature_types: List["FeatureType"],
    feature_values: List[Union[pd.Series, np.ndarray]],
) -> List[
    Union[
        "IntegerFeature",
        "FloatFeature",
        "StringFeature",
        "CategoricalFeature",
        "DatetimeFeature",
    ]
]:
    """The functuon takes three arguments. All arguments are lists and need to have an
    equal length.
    The function returns a list of features which will be used as an argument in the
    Validator class.

    Args:
        feature_names (List[str]): List with feature names.
        feature_types (List[FeatureType]): List with feature types.
        feature_values (List[Union[pd.Series, np.ndarray]]): List with feature values

    Raises:
        ValueError: Is raised if the input arguments have different lengths.
        NotImplementedError: Is raised if feature types are unknown.

    Returns:
        [List]: A list with Features.
    """
    if (len(feature_names) != len(feature_types)) | (
        len(feature_names) != len(feature_values)
    ):
        raise ValueError(
            (
                "The input arguments have inconsistent lengths:"
                f" feature_names={len(feature_names)},"
                f" feature_types={len(feature_types)},"
                f" feature_values={len(feature_values)}."
            )
        )

    features: List[
        Union[
            "IntegerFeature",
            "FloatFeature",
            "StringFeature",
            "CategoricalFeature",
            "DatetimeFeature",
        ]
    ] = []

    for f_name, f_type, f_values in zip(feature_names, feature_types, feature_values):

        if f_type == FeatureType.INT:
            feature = IntegerFeature(f_name)
        elif f_type == FeatureType.FLOAT:
            feature = FloatFeature(f_name)
        elif f_type == FeatureType.STRING:
            feature = StringFeature(f_name)
        elif f_type == FeatureType.CATEGORY:
            feature = CategoricalFeature(f_name)
        elif f_type == FeatureType.DATETIME:
            feature = DatetimeFeature(f_name)
        else:
            raise NotImplementedError(
                f"The feature type '{f_type}' is not implemented."
            )

        feature.describe_func(f_values)
        features.append(feature)

    return features


@dataclass
class Feature(ABC):
    """The Feature class defines the interface for the default and
    future customised feature validators."""

    name: str
    feature_type: "FeatureType"
    meta_data: Optional[Dict[str, Any]] = None

    @abstractmethod
    def describe_func(self, *args) -> None:
        """This method is used to update the meta_data dictionary
        which later will be used in its validate_func()."""

    @abstractmethod
    def validate_func(self, v: Any) -> None:
        """A method which performs validation accordently to the values
        in the meta_data dictionary. It returns the input value if valid
        otherwise raises a proper exepction."""

    def get_validator(self) -> pydantic.validator:
        """A method used to return the validator of a given feature."""
        val_args: Dict = {"pre": True, "each_item": True, "allow_reuse": True}

        if self.meta_data:
            return pydantic.validator(self.name, **val_args)(self.validate_func)
        else:
            return None


@dataclass
class IntegerFeature(Feature):
    """The default class used for an integer feature."""

    feature_type: "FeatureType" = FeatureType.INT

    def describe_func(
        self, x: Union[pd.Series, np.ndarray, list], x_percentile: float = 0.95
    ) -> None:
        """This method is used to update the meta_data dictionary for the
        default integer feature validation.

        Args:
            x (pd.Series, np.ndarray, list): A list of numbers.
            x_percentile (float, optional): The percentile we want to use as a
            warning level. Defaults to 0.95.
        """
        if not isinstance(x, (pd.Series, np.ndarray, list)):
            raise TypeError(
                f"The input argument 'x' must be a pandas Series, numpy array or list."
                f" The input argument is of type {type(x)}."
            )

        warn_low, warn_high = np.percentile(
            x,
            q=np.array(
                # compute lower and upper quantile
                [((1 - x_percentile) / 2) * 100, (1 - (1 - x_percentile) / 2) * 100]
            ),
        )

        self.meta_data = {
            "warn_low": warn_low,
            "warn_high": warn_high,
            "skip_low": min(x),
            "skip_high": max(x),
        }

    def validate_func(self, v: Union[int, float]) -> Union[int, float]:
        """The integer validation function which uses its meta_data dictionary.

        Args:
            v (Union[int, float]): An input value the model needs to validate.

        Raises:
            ValueError: Will raised if the input value does not comply with
            the validation schema.

        Returns:
            Union[int, float]: A valid input value.
        """
        skip_low: float = self.meta_data.get("skip_low", None)
        skip_high: float = self.meta_data.get("skip_high", None)
        warn_low: float = self.meta_data.get("warn_low", None)
        warn_high: float = self.meta_data.get("warn_high", None)

        if v < skip_low:
            raise FeatureValueError(
                message=(
                    f"{self.name}={v} - lower than skip level ({v} < {skip_low})."
                ),
                state=ValidatorStatus.NOT_VALID,
            )
        if v > skip_high:
            raise FeatureValueError(
                message=(
                    f"{self.name}={v} - higher than skip level ({v} > {skip_high})."
                ),
                state=ValidatorStatus.NOT_VALID,
            )
        if v < warn_low:
            raise FeatureValueError(
                message=f"{self.name}={v} - lower than warn level ({v} < {warn_low}).",
                state=ValidatorStatus.WARNING,
            )
        if v > warn_high:
            raise FeatureValueError(
                message=(
                    f"{self.name}={v} - higher than warn level ({v} > {warn_high})."
                ),
                state=ValidatorStatus.WARNING,
            )
        return v


@dataclass
class FloatFeature(IntegerFeature):
    """The default class used for an float feature."""

    feature_type: FeatureType = FeatureType.FLOAT


@dataclass
class StringFeature(Feature):
    """The default class used for an string (/character) feature."""

    feature_type: FeatureType = FeatureType.STRING

    def describe_func(self, x: Union[pd.Series, np.ndarray]) -> None:
        """A method to update the meta_data dictionary with unique string values.

        Args:
            x (pd.Series): A pandas series with object aka. characters
            (or categorical) values.
        """
        self.meta_data = {
            "unique_values": x.unique().tolist(),
        }

    def validate_func(self, v: str) -> str:
        """The string validation function which uses its meta_data dictionary.

        Args:
            v str: An input value the model needs to validate.

        Raises:
            ValueError: Will raised if the input value does not comply with
            the validation schema.

        Returns:
            str: A valid input value.
        """
        unique_values: List[str] = self.meta_data.get("unique_values", None)

        if v not in unique_values:
            raise FeatureValueError(
                message=f"{self.name}={v} - value is not in ({unique_values}).",
                state=ValidatorStatus.NOT_VALID,
            )
        return v


@dataclass
class CategoricalFeature(StringFeature):
    """The default class used for an categorical feature."""

    ordered: bool = False
    feature_type: FeatureType = FeatureType.CATEGORY


@dataclass
class DatetimeFeature(Feature):
    """The default class used for an integer feature."""

    feature_type: FeatureType = FeatureType.DATETIME
    datetime_format: str = "%Y-%m-%d %H:%M:%S.%f"

    def describe_func(self, x: Union[pd.Series, np.ndarray]) -> None:
        """This method is used to update the meta_data dictionary for the
        default datetime feature validation.

        Args:
            x (pd.Series): A pandas series with datetime values.
        """
        self.meta_data = {
            "skip_low": x.min(),
            "skip_high": x.max(),
        }

    def validate_func(self, v: datetime) -> datetime:
        """The datetime validation function which uses its meta_data dictionary.

        Args:
            v datetime: An input value the model needs to validate.

        Raises:
            ValueError: Will raised if the input value does not comply with
            the validation schema.

        Returns:
            datetime: A valid input value.
        """

        """"""
        skip_low: datetime = self.meta_data.get("skip_low", None)
        skip_high: datetime = self.meta_data.get("skip_high", None)

        if isinstance(v, str):
            v = datetime.strptime(v, self.datetime_format)
        if v < skip_low:
            raise FeatureValueError(
                message=f"{self.name}={v} - lower than skip level ({v} < {skip_low}).",
                state=ValidatorStatus.NOT_VALID,
            )
        if v > skip_high:
            raise FeatureValueError(
                message=(
                    f"{self.name}={v} - higher than skip level ({v} > {skip_high})."
                ),
                state=ValidatorStatus.NOT_VALID,
            )
        return v



import requests


def test__deployment__model_predict(deployment_base_url):
    """
    Invokes prediction on a default ASGI webapp using the provided URL.

    :param str deployment_base_url:
    """

    # -- Act -----------------------------------------------------------------

    response = requests.get(url=f"{deployment_base_url}")

    # -- Assert --------------------------------------------------------------

    assert response.status_code == 200
    assert response.json()["message"] == "Hello World"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" CLI module
"""

import click
from dotnetcore2 import runtime

# Hack OS version to avoid licencing problems in azureml
# TODO How else to get azureml Datasets working?
# TODO Move to somewhere else?
runtime.version = ("18", "10", "0")

from energinetml.settings import PACKAGE_VERSION  # noqa: E402

from .cluster import cluster_group  # noqa: E402
from .infrastructure import infrastructure_group  # noqa: E402
from .model import model_group  # noqa: E402
from .project import project_group  # noqa: E402
from .webapp import webapp_group  # noqa: E402


@click.command()
def version():
    """
    Prints SDK version.
    """
    click.echo(PACKAGE_VERSION)


@click.group()
def main():
    """
    Click main entrypoint.
    """
    pass


main.add_command(project_group, "project")
main.add_command(model_group, "model")
main.add_command(cluster_group, "cluster")
main.add_command(webapp_group, "webapp")
main.add_command(infrastructure_group, "infrastructure")
main.add_command(version)



"""
/predict request body JSON example:

    {
        "inputs": [
            {
                "identifier": "foo",
                "features": {
                    "age": 20,
                    "height": 180
                }
            },
            {
                "identifier": "bar",
                "features": {
                    "age": 30,
                    "height": 200
                }
            },
            {
                "identifier": "foo",
                "features": {
                    "age": 20,
                    "height": 180
                }
            }
        ]
    }

"""
import json
import time
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import pandas as pd
import pydantic
from opentelemetry import trace
from opentelemetry.trace import SpanKind

from energinetml.core.validation import ValidatorResponse, ValidatorStatus

if TYPE_CHECKING:
    from energinetml.core.model import Model, TrainedModel

tracer = trace.get_tracer(__name__)


class PredictionInput(list):
    """A class to use when handling data in :func:`~energinetml.Model.predict`"""

    def __init__(self, features: List[str], *args, **kwargs):
        """[summary]

        Args:
            features (List[str]): [description]
        """
        super().__init__(*args, **kwargs)
        self.features = features

    def as_dict_of_lists(self) -> Dict[str, Any]:
        """[summary]"""
        return {
            feature: [input[feature] for input in self] for feature in self.features
        }

    def as_pandas_dataframe(self) -> pd.DataFrame:
        """[summary]

        Raises:
            RuntimeError: [description]

        Returns:
            pd.DataFrame: [description]
        """

        return pd.DataFrame(self.as_dict_of_lists())


# -- Data models -------------------------------------------------------------


class PredictRequest(pydantic.BaseModel):
    """[summary]"""

    inputs: List[Any]

    def group_input_by_identifier(self):
        """[summary]

        Returns:
            [type]: TODO: I need help for this data structure.
        """
        inputs_per_identifier = {}

        for index, input in enumerate(self.inputs):
            if hasattr(input, "identifier"):
                identifier = input.identifier.value
            else:
                identifier = None

            inputs_per_identifier.setdefault(identifier, []).append(
                (index, dict(input.features))
            )

        return inputs_per_identifier.items()


class PredictResponse(pydantic.BaseModel):
    """[summary]"""

    predictions: List[Any]
    validations: Optional[List[Any]]


# -- Controller --------------------------------------------------------------


class PredictionController:
    """[summary]"""

    def __init__(
        self,
        model: "Model",
        trained_model: "TrainedModel",
        model_version: str = None,
    ):
        """[summary]

        Args:
            model (Model): [description]
            trained_model (TrainedModel): [description]
            model_version (str, optional): [description]. Defaults to None.
        """
        self.model = model
        self.trained_model = trained_model
        self.model_version = model_version

        self.predict_features = self.generate_validator_model(
            model_name="PredictFeatures", use_validation=False
        )
        self.predict_features_validator = self.generate_validator_model(
            model_name="PredictFeaturesValidator", use_validation=True
        )

    def generate_validator_model(
        self, model_name: str, use_validation: bool = True
    ) -> Optional[pydantic.BaseModel]:
        """A method for creating the validator object during initialization of the
        endpoint. The method depends on self.trained_model.validator and returns a
        pydantic.BaseModel if validator has been provided during training.

        Args:
            model_name (str): Name of the pydantic.BaseModel.
            use_validation (bool, optional): A boolean which enables the
            pydantic.BaseModel to use a validtor. Defaults to True.

        Returns:
            pydantic.BaseModel: A pydantic.BaseModel with or without a valdator.
        """
        validator = self.trained_model.validator

        if validator:
            validator_model = validator.get_pydantic_model(
                model_name=model_name, use_validation=use_validation
            )
        else:
            validator_model = None

        return validator_model

    @property
    def identifiers(self) -> List[str]:
        """[summary]

        Returns:
            List[str]: [description]
        """
        return self.trained_model.identifiers

    @property
    def features(self) -> List[str]:
        """[summary]

        Returns:
            List[str]: [description]
        """
        return self.trained_model.features

    @property
    def requires_identity(self) -> bool:
        """[summary]

        Returns:
            bool: [description]
        """
        return not self.trained_model.has_default_model()

    @cached_property
    def request_model(self):
        """
        Build a request model class (inherited from pydantic.BaseModel)
        based on a specific TrainedModel instance. The resulting model
        can be used for JSON input validation, Swagger docs etc.
        """

        predict_features = (
            self.predict_features
            if self.predict_features
            else pydantic.create_model(
                "PredictFeatures", **{feature: (Any, ...) for feature in self.features}
            )
        )

        predict_input_attributes = {"features": (predict_features, ...)}

        if self.requires_identity:
            identifier_enum = Enum("IdentifierEnum", {i: i for i in self.identifiers})
            predict_input_attributes["identifier"] = (identifier_enum, ...)

        predict_input = pydantic.create_model(
            "PredictInput", **predict_input_attributes
        )

        return pydantic.create_model(
            "PredictRequest",
            __base__=PredictRequest,
            inputs=(List[predict_input], ...),
        )

    @property
    def response_model(self):
        """[summary]"""
        return PredictResponse

    def predict(
        self, request: PredictRequest, correlation_id: str = None
    ) -> PredictResponse:
        """[summary]

        Args:
            request (PredictRequest): [description]
            correlation_id (str, optional): [description]. Defaults to None.

        Returns:
            PredictResponse: [description]
        """
        start_span = tracer.start_span(name="prediction", kind=SpanKind.SERVER)

        with start_span as span:
            start = time.perf_counter()
            identifiers, features, predictions, validations = self.invoke_model(request)
            end = time.perf_counter()

            if correlation_id:
                span.set_attribute("correlation_id", correlation_id)
            span.set_attribute("duration_model", str(end - start))
            span.set_attribute("model_name", self.model.name)
            if self.model_version is not None:
                span.set_attribute("model_version", self.model_version)
            span.set_attribute("identifiers", json.dumps(identifiers))
            span.set_attribute("features", json.dumps(features))
            span.set_attribute("predictions", json.dumps(predictions))
            span.set_attribute("validations", json.dumps(validations))

        return self.response_model(predictions=predictions, validations=validations)

    def _validate_features(self, features: Dict[str, Any]) -> ValidatorResponse:
        """validate method for a single entity.

        Args:
            features (Dict[str, Any]): Key value mapping of given features.

        Returns:
            ValidatorResponse: The validator response.
        """
        try:
            self.predict_features_validator(**features)
        except pydantic.ValidationError as pve:

            error_dict = json.loads(pve.json())
            error_states = [
                ValidatorStatus(item.get("ctx", {}).get("state")) for item in error_dict
            ]
            error_state = (
                ValidatorStatus.NOT_VALID
                if ValidatorStatus.NOT_VALID in error_states
                else ValidatorStatus.WARNING
            )

            return ValidatorResponse(
                state=error_state.name,
                message=json.dumps(error_dict),
            )

        else:
            return ValidatorResponse(state=ValidatorStatus.VALID.name)

    def validate_input_data(
        self, input_data: PredictionInput
    ) -> List[ValidatorResponse]:
        """An nput data validation method.

        Args:
            input_data (PredictionInput): Retrived input data

        Returns:
            List[ValidatorResponse]: A list with ValidatorReponses.
        """

        if self.predict_features_validator:

            validated_inputs: List[ValidatorResponse] = [
                self._validate_features(features) for features in input_data
            ]

        else:
            validated_inputs: List[ValidatorResponse] = [
                ValidatorResponse(state=ValidatorStatus.NOT_USED.name)
                for _ in input_data
            ]

        return validated_inputs

    def invoke_model(
        self, request: PredictRequest
    ) -> Tuple[List[str], List[str], List[Any], List[ValidatorResponse]]:
        """[summary]

        Args:
            request (PredictRequest): [description]

        Raises:
            RuntimeError: [description]

        Returns:
            PredictResponse: [description]
        """
        groups = list(request.group_input_by_identifier())
        identifiers_ordered = [... for _ in range(len(request.inputs))]
        features_ordered = [i for i in request.inputs]
        predictions_ordered = [... for _ in range(len(request.inputs))]
        validated_features_ordered: List[ValidatorResponse] = [
            ... for _ in range(len(request.inputs))
        ]

        # Invoke predict() for each unique identifier
        for identifier, inputs in groups:
            indexes = [i[0] for i in inputs]
            feature_sets = [i[1] for i in inputs]
            input_data = PredictionInput(self.features, feature_sets)

            input_data_validated = self.validate_input_data(input_data)
            predict_result = self.model.predict(
                trained_model=self.trained_model,
                identifier=identifier,
                input_data=input_data,
            )

            for index, features, prediction, validated_inputs in zip(
                indexes, feature_sets, predict_result, input_data_validated
            ):

                identifiers_ordered[index] = identifier
                features_ordered[index] = features
                predictions_ordered[index] = prediction
                if self.predict_features_validator:
                    validated_features_ordered[index] = validated_inputs.dict()

        # If elipsis still exist in the predictions,
        # there is not the same number of inputs as outputs.
        # This is will most likely be due to a wrong predict
        # function implementation
        if ... in predictions_ordered:
            raise RuntimeError(
                "The number of inputs and outputs is not equal"
                " for the given model. Please check the predict function.\n"
            )

        return (
            identifiers_ordered,
            features_ordered,
            predictions_ordered,
            validated_features_ordered if self.predict_features_validator else None,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

from typing import List, Optional

import requirements
from packaging import version

from energinetml.settings import DEFAULT_ENCODING


class RequirementList(List):
    """[summary]"""

    @classmethod
    def from_file(cls, filepath: str) -> "RequirementList":
        """[summary]

        Args:
            filepath (str): [description]

        Returns:
            RequirementList: [description]
        """
        with open(filepath, encoding=DEFAULT_ENCODING) as f:
            return cls(requirements.parse(f.read()))

    def __contains__(self, package_name: str) -> bool:
        """[summary]

        Args:
            package_name (str): [description]

        Returns:
            bool: [description]
        """
        return self.get(package_name) is not None

    def get(self, package_name: str):
        """[summary]

        Args:
            package_name (str): [description]

        Returns:
            TODO: Requirement or RequirementList: [description]
        """
        for requirement in self:
            if requirement.name == package_name:
                return requirement

    def get_specs(self, package_name: str) -> List[Optional[str]]:
        """[summary]

        Args:
            package_name (str): [description]

        Returns:
            List[Optional[str]]: [description]
        """
        requirement = self.get(package_name)
        if requirement is not None:
            return requirement.specs

    def get_version(self, package_name: str) -> Optional[version.Version]:
        """[summary]

        Args:
            package_name (str): [description]

        Returns:
            Optional[version.Version]: [description]
        """
        specs = self.get_specs(package_name)
        if specs:
            return version.parse(specs[0][1])



import os
import tempfile
from pathlib import Path
from unittest.mock import ANY, patch

import click
import pytest

from energinetml.cli.infrastructure.init import (
    _parse_input_path,
    _parse_input_project_name,
    _parse_input_resource_group,
    _parse_input_service_connection,
)

NAME = "NAME"
SUBSCRIPTION_NAME = "MY-SUBSCRIPTION"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"


# -- _parse_input_path() Tests -----------------------------------------------


@patch("energinetml.cli.infrastructure.init.click.prompt")
def test__parse_input_path__value_is_none__should_prompt_for_path(prompt_mock):
    """
    :param Mock prompt_mock:
    """

    # Act
    _parse_input_path(ctx=None, param=None, value=None)

    # Assert
    prompt_mock.assert_called_once_with(
        text="Enter project location", default=ANY, type=ANY
    )


@patch("energinetml.cli.infrastructure.init.click.echo")
def test__parse_input_path__path_is_a_file__should_abort_with_error(echo_mock):
    """
    :param Mock echo_mock:
    """
    with tempfile.TemporaryDirectory() as path:
        fp = os.path.join(path, "somefile.txt")

        Path(fp).touch()

        # Act + Assert
        with pytest.raises(click.Abort):
            _parse_input_path(ctx=None, param=None, value=fp)


@patch("energinetml.cli.infrastructure.init.click.prompt")
def test__parse_input_path__value_is_valid__should_return_value(prompt_mock):
    """
    :param Mock prompt_mock:
    """

    # Act
    returned_value = _parse_input_path(ctx=None, param=None, value="valid-value")

    # Assert
    assert returned_value == "valid-value"
    prompt_mock.assert_not_called()


# -- _parse_input_project_name() Tests ---------------------------------------


@patch("energinetml.cli.infrastructure.init.click.prompt")
def test__parse_input_project_name__should_prompt_for_value_and_return_it(prompt_mock):
    """
    :param Mock prompt_mock:
    """

    # Arrange
    prompt_mock.return_value = "projectname"

    # Act
    returned_value = _parse_input_project_name(ctx=None, param=None, value=None)

    # Assert
    assert returned_value == "projectname"

    prompt_mock.assert_called_once_with(text="Please enter a project name", type=ANY)


# -- _parse_input_resource_group() Tests -------------------------------------


@patch("energinetml.cli.infrastructure.init.click.prompt")
def test__parse_input_resource_group__should_prompt_for_value_and_return_it(
    prompt_mock,
):  # noqa: E501
    """
    :param Mock prompt_mock:
    """

    # Arrange
    prompt_mock.return_value = "rgname"

    # Act
    returned_value = _parse_input_resource_group(ctx=None, param=None, value=None)

    # Assert
    assert returned_value == "rgname"

    prompt_mock.assert_called_once_with(text="Enter resource group name", type=ANY)


# -- _parse_input_service_connection() Tests ---------------------------------


@patch("energinetml.cli.infrastructure.init.click.prompt")
def test__parse_input_service_connection__should_prompt_for_value_and_return_it(
    prompt_mock,
):  # noqa: E501
    """
    :param Mock prompt_mock:
    """

    # Arrange
    prompt_mock.return_value = "scname"

    # Act
    returned_value = _parse_input_service_connection(ctx=None, param=None, value=None)

    # Assert
    assert returned_value == "scname"

    prompt_mock.assert_called_once_with(
        text="Enter Azure DevOps service connection name", type=ANY
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click

from energinetml.cli.utils import discover_project
from energinetml.core.project import WebAppProject


@click.command()
@discover_project(WebAppProject)
@click.option(
    "--host",
    default="127.0.0.1",
    type=str,
    help="Host to serve on (default: 127.0.0.1)",
)
@click.option("--port", default=8080, type=int, help="Port to serve on (default: 8080)")
def serve(host: str, port: int, project: WebAppProject) -> None:
    """Serve a HTTP web API for model prediction."""
    engine = project.get_engine()
    engine.serve(project)



"""[summary]
"""
import json
import os
from typing import TYPE_CHECKING, Any, Dict, List

import azureml
from azureml.core import Experiment, Run

from energinetml.azure.datasets import (
    DownloadedAzureMLDataStore,
    MountedAzureMLDataStore,
)
from energinetml.azure.logger import AzureMlLogger
from energinetml.core.logger import ConsoleLogger
from energinetml.core.training import AbstractTrainingContext
from energinetml.settings import DEFAULT_RELATIVE_ARTIFACT_PATH

if TYPE_CHECKING:
    from energinetml.azure.backend import AzureBackend
    from energinetml.core.model import Model, TrainedModel


class AzureTrainingContext(AbstractTrainingContext):
    def save_artifacts(self, model: "Model"):
        """Save model artifacts to cloud.

        Args:
            model (Model): [description]

        Raises:
            AbstractBackend.BackendException: [description]
        """
        try:
            self.az_run.upload_folder(
                name=DEFAULT_RELATIVE_ARTIFACT_PATH, path=model.artifact_path
            )
        except azureml._common.exceptions.AzureMLException as ex:
            raise self.backend.parse_azureml_exception(ex)

    def save_logs(self, *cloggers: List[ConsoleLogger]):
        try:
            for clogger in cloggers:
                clogger.flush()
                self.az_run.upload_file(clogger.filename, clogger.filepath)
        except azureml._common.exceptions.AzureMLException as ex:
            raise self.backend.parse_azureml_exception(ex)

    def save_meta_data(self, meta_data: Dict[str, str], meta_file_path: str):
        """Create a meta json-file with the content from a meta_data dictionary
        and save it to the path specified.

        Args:
            meta_data (Dict[str, str])): A dictionary of meta_data to persist in the
            json-file
            meta_file_path (str): The local path where the json-file is persisted.

        """
        json_object = json.dumps(meta_data, indent=4)
        with open(meta_file_path, "w") as outfile:
            outfile.write(json_object)

    def get_portal_url(self) -> str:
        """Retrieve the URL to the AzureML Run.

        Raises:
            AbstractBackend.BackendException: [description]

        Returns:
            str:
        """
        try:
            return self.az_run.get_portal_url()
        except azureml._common.exceptions.AzureMLException as ex:
            raise self.backend.parse_azureml_exception(ex)


class AzureLocalTrainingContext(AzureTrainingContext):
    """[summary]"""

    def __init__(self, backend: "AzureBackend", force_download: bool):
        """[summary]

        Args:
            backend (AzureBackend): [description]
            force_download (bool): [description]
        """
        self.backend = backend
        self.force_download = force_download
        self.az_run = None

    def train_model(
        self, model: "Model", tags: Dict[str, Any], **kwargs
    ) -> "TrainedModel":
        """[summary]

        Args:
            model (Model): [description]
            tags (Dict[str, Any]): [description]

        Raises:
            AbstractBackend.BackendException: [description]
            AbstractBackend.BackendException: [description]

        Returns:
            TrainedModel: [description]
        """
        project_meta = model.project.as_dict()
        az_workspace = self.backend.get_workspace(project_meta)

        az_experiment = Experiment(workspace=az_workspace, name=model.experiment)

        datasets = DownloadedAzureMLDataStore.from_model(
            model=model,
            datasets=model.datasets_parsed.local,
            workspace=az_workspace,
            force_download=self.force_download,
        )

        with model.temporary_folder() as temp_path:
            # The "outputs" parameter is provided here with a non-existing
            # folder path to avoid having azureml upload files. We will do
            # this manually somewhere else.

            try:
                self.az_run = az_experiment.start_logging(
                    snapshot_directory=temp_path,
                    outputs=os.path.join(temp_path, "a-folder-that-does-not-exists"),
                )
            except azureml._common.exceptions.AzureMLException as ex:
                raise self.backend.parse_azureml_exception(ex)

            self.az_run.set_tags(tags)
            try:
                return model.train(
                    datasets=datasets, logger=AzureMlLogger(self.az_run), **kwargs
                )
            finally:
                try:
                    self.az_run.complete()
                except azureml._common.exceptions.AzureMLException as ex:
                    raise self.backend.parse_azureml_exception(ex)

    def get_parameters(self, model: "Model") -> Dict:
        """[summary]

        Args:
            model (Model): [description]

        Returns:
            Dict: [description]
        """
        params = {}
        params.update(model.parameters)
        params.update(model.parameters_local)
        return params

    def get_tags(self, model: "Model") -> Dict:
        """[summary]

        Args:
            model (Model): [description]

        Returns:
            Dict: [description]
        """
        return {"datasets": ", ".join(model.datasets + model.datasets_local)}


class AzureCloudTrainingContext(AzureTrainingContext):
    """[summary]"""

    def __init__(self):
        self.az_run = None

    def train_model(
        self, model: "Model", tags: Dict[str, Any], **kwargs
    ) -> "TrainedModel":
        """[summary]

        Args:
            model (Model): [description]
            tags (Dict[str, Any]): [description]

        Returns:
            TrainedModel: [description]
        """
        self.az_run = Run.get_context(allow_offline=False)
        self.az_run.set_tags(tags)

        datasets = MountedAzureMLDataStore.from_model(
            model=model,
            datasets=model.datasets_parsed.cloud,
            workspace=self.az_run.experiment.workspace,
        )

        try:
            return model.train(
                datasets=datasets, logger=AzureMlLogger(self.az_run), **kwargs
            )
        finally:
            self.az_run.complete()

    def get_parameters(self, model: "Model") -> Dict:
        """[summary]

        Args:
            model (Model): [description]

        Returns:
            Dict: [description]
        """
        params = {}
        params.update(model.parameters)
        params.update(model.parameters_cloud)
        return params

    def get_tags(self, model: "Model") -> Dict:
        """[summary]

        Args:
            model (Model): [description]

        Returns:
            Dict: [description]
        """
        return {"datasets": ", ".join(model.datasets + model.datasets_cloud)}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import discover_model
from energinetml.cli.utils.verify import verify_compute_cluster
from energinetml.core.model import Model


@click.command()
@discover_model()
@click.option(
    "--cluster-name",
    "cluster_name",
    required=False,
    default=None,
    type=str,
    help="Name of compute cluster, if creating a new cluster",
)
def change(model: Model, cluster_name: str) -> None:
    """Switch to use another (existing) compute cluster."""

    verify_compute_cluster(backend, model, cluster_name, modify_vm_size=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import json
import os
import subprocess
import sys
from typing import Dict

from energinetml.core.model import Model
from energinetml.core.project import WebAppProject
from energinetml.settings import (
    DEFAULT_RELATIVE_ARTIFACT_PATH,
    DOCKERFILE_PATH_ML_MODEL,
    PACKAGE_VERSION,
)


def build_prediction_api_docker_image(
    artifact_path: str, tag: str, model_version: str
) -> None:
    """TODO: Add package version when installing energinet-ml-sdk

    Args:
        path (str): [description]
        tag (str): [description]
        model_version (str): [description]

    Raises:
        ValueError: [description]
    """
    meta = _metadata_from_artifact(artifact_path)
    # Convert module path to folder path
    model_path = meta["module_name"].replace(".", "/")

    build_docker_image(
        path=artifact_path,
        tag=tag,
        dockerfile_path=DOCKERFILE_PATH_ML_MODEL,
        build_args={
            "MODEL_PATH": model_path,
            "OUTPUTS_PATH": DEFAULT_RELATIVE_ARTIFACT_PATH,
            "MODEL_VERSION": model_version,
        },
    )


def _metadata_from_artifact(artifact_path: str) -> Dict:
    """Read a meta file.

    Args:
        artifact_path (str): [description]

    Returns:
        Dict: [description]
    """
    meta_file_path = os.path.join(
        artifact_path,
        DEFAULT_RELATIVE_ARTIFACT_PATH,
        Model._META_FILE_NAME,
    )

    with open(meta_file_path, "r") as f:
        meta = json.load(f)

    return meta


def build_webapp_docker_image(project: WebAppProject, tag: str) -> None:
    """TODO: Add package version when installing energinet-ml-sdk

    Args:
        project (project.WebAppProject): [description]
        tag (str): [description]
    """
    build_docker_image(
        path=project.path, tag=tag, dockerfile_path=project.dockerfile_path
    )


def build_docker_image(
    path: str,
    tag: str,
    params: Dict[str, str] = None,
    build_args: Dict[str, str] = None,
    dockerfile_path: str = None,
):
    """Build a Docker image.

    Args:
        path (str): [description]
        tag (str): [description]
        params (Dict[str, str], optional): [description]. Defaults to None.
        build_args (Dict[str, str], optional): [description]. Defaults to None.
        dockerfile_path (str, optional): [description]. Defaults to None.
    """
    if params is None:
        params = {}
    if build_args is None:
        build_args = {}

    if dockerfile_path:
        params["--file"] = dockerfile_path

    build_args["PACKAGE_VERSION"] = str(PACKAGE_VERSION)

    # Render 'docker build' command
    command = ["docker", "build"]
    command.extend(("--tag", tag))
    for k, v in params.items():
        command.extend((k, v))
    for k, v in build_args.items():
        command.extend(("--build-arg", f"{k}={v}"))
    command.append(path)

    # Run 'docker build' command in subprocess
    subprocess.check_call(
        command, stdout=sys.stdout, stderr=subprocess.STDOUT, shell=False
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage compute clusters for a model.
"""
import click

from .change import change as change_cluster
from .create import create as create_cluster


@click.group()
def cluster_group() -> None:
    """Manage compute clusters for a model."""
    pass


cluster_group.add_command(create_cluster)
cluster_group.add_command(change_cluster)



import json
from datetime import date, datetime
from unittest.mock import Mock, patch

from energinetml.core.insight import QUERY, parse_return_json, query, query_predictions


@patch("energinetml.core.insight.subprocess.check_output")
def test__query(check_output_mock):
    """
    :param Mock check_output_mock:
    """
    app = "APP"
    resource_group = "RESOURCE-GROUP"
    subscription = "SUBSCRIPTION"
    start_date = date(2020, 1, 1)
    end_date = date(2020, 12, 31)

    expected_return_json = b'{"some": "json"}'
    expected_command = [
        "az",
        "monitor",
        "app-insights",
        "query",
        "--apps",
        app,
        "--resource-group",
        resource_group,
        "--subscription",
        subscription,
        "--start-time",
        start_date.isoformat(),
        "--end-time",
        end_date.isoformat(),
        "--analytics-query",
        QUERY,
    ]

    check_output_mock.return_value = (
        expected_return_json + b"EXTRA CHARACTERS THAT SHOULD BE REMOVED"
    )

    # Act

    returned_json = query(
        app=app,
        resource_group=resource_group,
        subscription=subscription,
        start_date=start_date,
        end_date=end_date,
    )

    # Assert

    check_output_mock.assert_called_once_with(expected_command, shell=False)

    assert returned_json == json.loads(expected_return_json)


def test__parse_return_json():
    input_features = '[{"sepal_length": 1, "sepal_width": 2, "petal_length": 3, "petal_width": 4}, {"sepal_length": 10, "sepal_width": 20, "petal_length": 30, "petal_width": 40}]'  # noqa: E501
    model_name = "d-tree"
    predictions = '["Iris-versicolor", "Iris-virginica"]'
    indentifiers = "[null, null]"
    model_version = "456"

    return_json = {
        "tables": [
            {
                "columns": [
                    {"name": "timestamp", "type": "datetime"},
                    {"name": "identifiers", "type": "dynamic"},
                    {"name": "features", "type": "dynamic"},
                    {"name": "predictions", "type": "dynamic"},
                    {"name": "model_name", "type": "dynamic"},
                    {"name": "model_version", "type": "dynamic"},
                ],
                "name": "PrimaryResult",
                "rows": [
                    [
                        "2021-03-15T11:15:55.986Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:55.854Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:55.709Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:55.563Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:55.463Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:52.904Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:52.739Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:15:40.224Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                    [
                        "2021-03-15T11:13:49.717Z",
                        indentifiers,
                        input_features,
                        predictions,
                        model_name,
                        model_version,
                    ],
                ],
            }
        ]
    }

    # Act

    result = list(parse_return_json(return_json))

    # Assert

    assert result == [
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 55),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 52),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 52),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 52),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 52),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 40),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 15, 40),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 13, 49),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 1,
                "sepal_width": 2,
                "petal_length": 3,
                "petal_width": 4,
            },
            "prediction": "Iris-versicolor",
        },
        {
            "timestamp": datetime(2021, 3, 15, 11, 13, 49),
            "model_name": model_name,
            "model_version": model_version,
            "identifier": None,
            "features": {
                "sepal_length": 10,
                "sepal_width": 20,
                "petal_length": 30,
                "petal_width": 40,
            },
            "prediction": "Iris-virginica",
        },
    ]


@patch("energinetml.core.insight.query")
@patch("energinetml.core.insight.parse_return_json")
def test__query_predictions(parse_return_json_mock, query_mock):
    """
    :param Mock parse_return_json_mock:
    :param Mock query_mock:
    """
    query_return_value = Mock()
    parse_return_json_return_value = Mock()

    query_mock.return_value = query_return_value
    parse_return_json_mock.return_value = parse_return_json_return_value

    # Act
    return_value = query_predictions(some="args", foobar=1234)

    # Assert
    query_mock.assert_called_once_with(some="args", foobar=1234)
    parse_return_json_mock.assert_called_once_with(query_return_value)
    assert return_value is parse_return_json_return_value



import os
import tempfile

import pytest
from packaging.version import Version

from energinetml.core.requirements import RequirementList

REQUIREMENTS = ["some-package==1.0.0", "another-package"]


@pytest.fixture
def requirements_list():
    with tempfile.TemporaryDirectory() as path:
        fp = os.path.join(path, "requirements.txt")

        with open(fp, "w") as f:
            f.write("\n".join(REQUIREMENTS))

        yield RequirementList.from_file(fp)


class TestRequirementList:
    def test__contains(self, requirements_list):
        """
        :param RequirementList requirements_list:
        """
        assert "some-package" in requirements_list
        assert "another-package" in requirements_list
        assert "john-doe" not in requirements_list

    @pytest.mark.parametrize("package", ("some-package", "another-package"))
    def test__get__requirement_exists__should_return_requirement_object(
        self, package, requirements_list
    ):
        """
        :param str package:
        :param RequirementList requirements_list:
        """
        requirement = requirements_list.get(package)

        assert hasattr(requirement, "name")
        assert hasattr(requirement, "specs")

    def test__get__requirement_does_not_exists__should_return_none(
        self, requirements_list
    ):
        """
        :param RequirementList requirements_list:
        """
        assert requirements_list.get("john-doe") is None

    @pytest.mark.parametrize(
        "package, specs", (("some-package", [("==", "1.0.0")]), ("another-package", []))
    )
    def test__get_specs__should_return_correct_specs(
        self, package, specs, requirements_list
    ):
        """
        :param str package:
        :param typing.List[typing.Optional[str]] specs:
        :param RequirementList requirements_list:
        """
        assert requirements_list.get_specs(package) == specs

    @pytest.mark.parametrize(
        "package, version",
        (("some-package", Version("1.0.0")), ("another-package", None)),
    )
    def test__get_version__should_return_correct_version(
        self, package, version, requirements_list
    ):
        """
        :param str package:
        :param typing.Optional[Version] version:
        :param RequirementList requirements_list:
        """
        assert requirements_list.get_version(package) == version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import json
import os

import click

from energinetml.cli.utils import discover_model
from energinetml.core.http import run_predict_api
from energinetml.core.model import Model
from energinetml.settings import DEFAULT_RELATIVE_ARTIFACT_PATH


@click.command()
@discover_model()
@click.option(
    "--outputs-path",
    "outputs_path",
    default=os.path.join(DEFAULT_RELATIVE_ARTIFACT_PATH),
    type=click.Path(file_okay=True, resolve_path=True),
    help="Path to the outputs directory where the model.pkl and metadata.json exist.",
)
@click.option(
    "--host",
    default="127.0.0.1",
    type=str,
    help="Host to serve on (default: 127.0.0.1)",
)
@click.option("--port", default=8080, type=int, help="Port to serve on (default: 8080)")
@click.option(
    "--model-version",
    "model_version",
    required=False,
    type=str,
    default="Unspecified",
    help="Model version (used for logging)",
)
def serve(
    host: str,
    port: int,
    model: Model,
    outputs_path: str,
    model_version: str = None,
) -> None:
    """Serve a HTTP web API for model prediction."""
    trained_model_path = os.path.join(outputs_path, model._TRAINED_MODEL_FILE_NAME)
    trained_model = model.load(trained_model_path)

    meta_data_path = os.path.join(outputs_path, model._META_FILE_NAME)
    with open(meta_data_path) as meta_data_file:
        meta_data = json.load(meta_data_file)

    run_predict_api(
        model=model,
        trained_model=trained_model,
        meta_data=meta_data,
        model_version=model_version,
        host=host,
        port=port,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from __future__ import annotations

import json
import subprocess
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Iterable  # noqa TYP001

if TYPE_CHECKING:
    import pandas as pd

QUERY = """
    requests
        | where name == "prediction"
        | order by timestamp
        | project timestamp,
                  identifiers = customDimensions["identifiers"],
                  features = customDimensions["features"],
                  predictions = customDimensions["predictions"],
                  model_name = customDimensions["model_name"],
                  model_version = customDimensions["model_version"]
        | where features != ""
"""


def query(
    app: str,
    resource_group: str,
    subscription: str,
    start_date: datetime.date,
    end_date: datetime.date,
) -> Dict[str, Any]:
    """Query Application Insights for model predictions.

    Args:
        app (str): Azure Application Insight app name
        resource_group (str): Azure resource group
        subscription (str): Azure subscription
        start_date (datetime.date): Start date (included)
        end_date (datetime.date): End date (included)

    Returns:
        Dict[str, Any]: [description]
    """
    command = ["az", "monitor", "app-insights", "query"]
    command.extend(("--apps", app))
    command.extend(("--resource-group", resource_group))
    command.extend(("--subscription", subscription))
    command.extend(("--start-time", start_date.isoformat()))
    command.extend(("--end-time", end_date.isoformat()))
    command.extend(("--analytics-query", QUERY))

    output_bytes = subprocess.check_output(command, shell=False)
    output_string = output_bytes.decode("utf-8")
    # String may contain trailing characters, not part of the JSON document
    output_string = output_string[: output_string.rfind("}") + 1]

    return json.loads(output_string)


def parse_return_json(return_json: Dict[str, Any]) -> Iterable[Dict[str, Any]]:
    """[summary]

    Args:
        return_json (Dict[str, Any]): [description]

    Returns:
        Iterable[Dict[str, Any]]: [description]

    Yields:
        Iterator[Iterable[Dict[str, Any]]]: [description]
    """
    columns = [c["name"] for c in return_json["tables"][0]["columns"]]

    for row in return_json["tables"][0]["rows"]:
        entity = dict(zip(columns, row))

        t = datetime.strptime(entity["timestamp"], "%Y-%m-%dT%H:%M:%S.%fZ").replace(
            microsecond=0
        )

        z = zip(
            json.loads(entity["identifiers"]),
            json.loads(entity["features"]),
            json.loads(entity["predictions"]),
        )

        for identifier, features, prediction in z:
            yield {
                "timestamp": t,
                "model_name": entity["model_name"],
                "model_version": entity["model_version"],
                "identifier": identifier,
                "features": features,
                "prediction": prediction,
            }


def query_predictions(**kwargs) -> Iterable[Dict[str, Any]]:
    """Perform an Application Insight query to get inputs with their respective
    predictions from a deployed model. These are the entries logged by models
    when deployed in Azure using the SDK.

    Example usage:

        from datetime import date
        from energinetml import query_predictions

        entries = query_predictions(
            app='MyApplicationInsightAppName',
            resource_group='MyAzureResourceGroup',
            subscription='MyAzureSubscription',
            start_date=date(2021, 1, 1),
            end_date=date(2021, 12, 31),
        )

    Returns:
        Iterable[Dict[str, Any]]: [description]
    """
    return parse_return_json(query(**kwargs))


def query_predictions_as_dataframe(**kwargs) -> pd.DataFrame:
    """Shortcut for query_predictions() but returns a Pandas DataFrame
    instead of a list of dicts.
    Any kwargs: Keyword args for query_predictions() method

    Raises:
        RuntimeError: [description]

    Returns:
        pd.DataFrame: [description]
    """

    try:
        import pandas as pd
    except ImportError:
        raise RuntimeError(
            "Failed to import pandas. Make sure to add "
            "pandas to your requirements.txt file"
        )

    return pd.DataFrame(query_predictions(**kwargs))



import os
import subprocess
import sys
from unittest.mock import patch

from energinetml.core.docker import build_prediction_api_docker_image
from energinetml.settings import (
    DEFAULT_RELATIVE_ARTIFACT_PATH,
    DOCKERFILE_PATH_ML_MODEL,
    PACKAGE_VERSION,
)


@patch("energinetml.core.docker._metadata_from_artifact")
@patch("energinetml.core.docker.subprocess.check_call")
def test__build_prediction_api_docker_image__happy_path(
    check_call_mock, _metadata_from_artifact_mock
):
    """
    :param Mock check_call_mock:
    """
    path = os.path.join("some", "arbitrary", "path")
    tag = "my-model:v1"
    model_version = "123"
    model_module_name = "myrepo.mymodel"

    _metadata_from_artifact_mock.return_value = {
        "module_name": model_module_name,
    }

    # Act

    build_prediction_api_docker_image(
        artifact_path=path,
        tag=tag,
        model_version=model_version,
    )

    # Assert

    expected_command = [
        "docker",
        "build",
        "--tag",
        tag,
        "--file",
        DOCKERFILE_PATH_ML_MODEL,
        "--build-arg",
        "MODEL_PATH=%s" % model_module_name.replace(".", "/"),
        "--build-arg",
        "OUTPUTS_PATH=%s" % DEFAULT_RELATIVE_ARTIFACT_PATH,
        "--build-arg",
        "MODEL_VERSION=%s" % model_version,
        "--build-arg",
        "PACKAGE_VERSION=%s" % PACKAGE_VERSION,
        path,
    ]

    check_call_mock.assert_called_once_with(
        expected_command, stdout=sys.stdout, stderr=subprocess.STDOUT, shell=False
    )



# Project
PROJECT_NAME = "NAME"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"
CLUSTER_NAME = "CLUSTER-NAME"


# Model
MODEL_NAME = "NAME"
EXPERIMENT = "EXPERIMENT"
COMPUTE_TARGET = "COMPUTE-TARGET"
VM_SIZE = "VM-SIZE"
DATASETS = ["iris", "hades:2"]
DATASETS_LOCAL = ["local:2"]
DATASETS_CLOUD = ["cloud:3"]
FEATURES = ["feature1", "feature2"]
PARAMETERS = {"param1": "value1", "param2": "value2"}
FILES_INCLUDE = ["__init__.py", "__main__.py", "model.json", "model.py"]

RUN_ID = "0123456789"
EXPERIMENT_NAME = "experiment-name"

# This satisfies SonarCloud
JSON_FILE_NAME = "something.json"
VALID_JSON = '{"valid": "json"}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The logger module is essential for logging an experiment important numbers
doing a training.
"""
from __future__ import annotations

import io
import os
import sys
from typing import TYPE_CHECKING, Any, Tuple  # noqa TYP001

if TYPE_CHECKING:
    import pandas as pd

from energinetml.settings import (
    DEFAULT_LOG_ENCODING,
    DEFAULT_LOG_FILENAME,
    DEFAULT_LOG_PATH,
)


class ConsoleLogger:
    """This class enables logging of stdout and stderr.
    The class is used during local trainings and finally uploaded to the azure ml
    expirment.

    Args:
        name (str): Name of the log file.
        console (io.TextIOWrapper, optional): Possible values
            [sys.stdout, sys.stderr]. Defaults to sys.stdout.

    Returns:
        object: A console logger object which can be used to overwrite
            the default sys.std* object.
    """

    def __init__(
        self,
        name: str,
        console: io.TextIOWrapper = sys.stdout,
    ):
        """Init ConsoleLogger with either sys.stdout or sys.stderr."""
        self.console = console

        file_prefix, ext = DEFAULT_LOG_FILENAME.split(".")
        self.filename = f"{file_prefix}_{name}.{ext}"

        self.filepath, self.log = self._init_log_file(self.filename)

    def _init_log_file(self, filename: str) -> Tuple[str, io.TextIOWrapper]:
        """Performs Initialization of the log file and further enrich the object."""
        os.makedirs(DEFAULT_LOG_PATH, exist_ok=True)

        filepath = f"{DEFAULT_LOG_PATH}/{filename}"
        return filepath, open(filepath, "w", encoding=DEFAULT_LOG_ENCODING)

    def isatty(self) -> bool:
        """A requried function for sys.stdout and sys.stderr.

        Returns:
            bool: This function will only return False.
        """
        return False

    def write(self, message: str) -> None:
        """A requried function for sys.stdout and sys.stderr.
        The function will both print to file and to console.

        Args:
            message (str): The string which needs to be appended to the log.
        """
        self.log.write(message)
        self.console.write(message)

    def flush(self):
        """The function flush the io buffer.
        This function is called before uploading to azure ml.
        """
        self.log.flush()


class MetricsLogger:
    """A logger to use for logging experiments."""

    def echo(self, s: str) -> None:
        """Write to console.

        Args:
            s: The string to be shown.

        Raises:
            NotImplementedError: [description]

        Example:
            >>> s = "Loading data"
            >>> logger.echo(s)

        """
        raise NotImplementedError

    def log(self, name: str, value: Any) -> None:
        """Log a name-value pair.

        Args:
            name (str): Name for the logged parameter.
            value (Any): Value for the logged parameter.

        Example:
            >>> name = "score"
            >>> value = model.score()
            >>> logger.log(name, value)

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def tag(self, key: str, value: str) -> None:
        """[summary]

        Args:
            key (str): [description]
            value (str): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError

    def dataframe(self, name: str, df: pd.DataFrame) -> None:
        """[summary]

        Args:
            name (str): [description]
            df (pd.DataFrame): [description]

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage Python Web Apps.
"""

import click

from .build import build as build_webapp
from .init import init as init_webapp
from .serve import serve as serve_webapp


@click.group()
def webapp_group():
    """Manage Python Web Apps."""
    pass


webapp_group.add_command(init_webapp)
webapp_group.add_command(build_webapp)
webapp_group.add_command(serve_webapp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import pprint
from typing import Any, List, Union

import azureml
import pandas

from energinetml.core.logger import MetricsLogger


class AzureMlLogger(MetricsLogger):
    """[summary]"""

    def __init__(self, run: azureml.core.Run):
        """[summary]

        Args:
            run (azureml.core.Run): [description]
        """
        self.run = run

    def echo(self, s: str) -> None:
        """Echo wrapper.

        Args:
            s (str): String we want to log.
        """
        print(s)

    def log(self, name: str, value: Union[float, int, str]) -> None:
        """[summary]

        Args:
            name (str): [description]
            value (Union[float, int, str]): [description]
        """
        self.run.log(name, value)
        self.echo(f"LOG: {name} = {value}")

    def tag(self, key: str, value: str) -> None:
        """[summary]

        Args:
            key (str): [description]
            value (str): [description]
        """
        self.run.tag(key, value)
        self.echo(f"TAG: {key} = {value}")

    def table(
        self, name: Union[str, Any], dict_of_lists: List[dict], echo=True
    ) -> None:
        """[summary]

        Args:
            name (Union[str, Any]): [description]
            dict_of_lists (List[dict]): [description]
            echo (bool, optional): [description]. Defaults to True.
        """
        list_of_dicts = [
            dict(zip(dict_of_lists, t)) for t in zip(*dict_of_lists.values())
        ]

        for _dict in list_of_dicts:
            self.run.log_table(name, _dict)

        if echo:
            # TODO print actual table
            self.echo(f"{name}:")
            self.echo(pprint.PrettyPrinter(indent=4).pformat(dict_of_lists))

    def dataframe(self, name: str, df: pandas.DataFrame) -> None:
        """[summary]

        Args:
            name (str): [description]
            df (pandas.DataFrame): [description]
        """
        df = df.reset_index()
        self.table(name, df.to_dict(orient="list"), echo=False)
        self.echo(df.to_string())



from unittest.mock import ANY, patch

from click.testing import CliRunner

from energinetml.cli.model.build import build

# -------------- Tests ----------------------------------------------------------


@patch("energinetml.cli.model.build.build_prediction_api_docker_image")
def test__build_model__should__build_docker_image(
    build_prediction_api_docker_image_mock, model_path
):
    """
    :param Mock build_prediction_api_docker_image_mock:
    """
    runner = CliRunner()

    # Act
    result = runner.invoke(
        cli=build,
        args=[
            "--artifact-path",
            model_path,
            "--tag",
            "docker:tag",
            "--model-version",
            "123",
        ],
    )

    # Assert
    assert result.exit_code == 0

    build_prediction_api_docker_image_mock.assert_called_once_with(
        artifact_path=ANY,  # Temporary directory
        model_version="123",
        tag="docker:tag",
    )



from unittest.mock import Mock

import numpy as np

from energinetml.core.model import TrainedModel
from energinetml.core.predicting import PredictionController
from energinetml.core.validation import (
    FeatureType,
    Validator,
    ValidatorStatus,
    create_feature_models,
)


class TestPredictionController:
    def test__predictions_returned_in_correct_order(self):
        def __model_predict(trained_model, identifier, input_data):
            """
            :param TrainedModel trained_model:
            :param str identifier:
            :param typing.List[typing.Dict[str, typing.Any]] input_data:
            """
            return [identifier for _ in range(len(input_data))]

        feature_names = ["feature1", "feature2"]
        feature_models = create_feature_models(
            feature_names=feature_names,
            feature_types=[FeatureType.INT, FeatureType.INT],
            feature_values=[np.array([1, 2, 3]), np.array([1, 2, 3])],
        )

        trained_model = TrainedModel(
            features=feature_names,
            models={"identifier1": Mock(), "identifier2": Mock()},
            validator=Validator(dataset_id="test", feature_models=feature_models),
        )
        uut = PredictionController(
            model=Mock(predict=__model_predict),
            trained_model=trained_model,
            model_version="123",
        )

        request = uut.request_model(
            inputs=[
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 1, "feature2": 1},
                },
                {
                    "identifier": "identifier2",
                    "features": {"feature1": 2, "feature2": 2},
                },
                {
                    "identifier": "identifier1",
                    "features": {"feature1": 30, "feature2": 3},
                },
            ]
        )

        # Act
        result = uut.predict(request, correlation_id="uuid")

        # Assert
        assert result.predictions == ["identifier1", "identifier2", "identifier1"]
        assert result.validations is not None
        assert result.validations[0].get("state") == ValidatorStatus.WARNING.name
        assert result.validations[1].get("state") == ValidatorStatus.VALID.name
        assert result.validations[-1].get("state") == ValidatorStatus.NOT_VALID.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import os
import sys
from typing import List, Union

import click
import click_spinner

from energinetml.backend import default_backend as backend
from energinetml.cli.utils import discover_model
from energinetml.core.logger import ConsoleLogger
from energinetml.core.model import Model, TrainedModel
from energinetml.settings import PACKAGE_NAME, PACKAGE_VERSION

# -- CLI Command -------------------------------------------------------------


@click.command()
@discover_model()
@click.argument("parameters", nargs=-1)
@click.option(
    "--cloud-mode",
    "-c",
    "cloud_mode",
    default=False,
    is_flag=True,
    help="Run training in cloud mode (do not use locally)",
)
@click.option(
    "--force-download",
    "-f",
    "force_download",
    default=False,
    is_flag=True,
    help="Force download of datasets (ignore locally cached files)",
)
@click.option("--seed", "-s", required=False, default=None, type=str, help="Seed value")
def train(
    parameters: List[str],
    cloud_mode: bool,
    force_download: bool,
    seed: Union[int, str],
    model: Model,
):
    """Train a model locally"""

    try:
        stdout_console_logger = ConsoleLogger(name="stdout", console=sys.stdout)
        sys.stdout = stdout_console_logger
        stderr_console_logger = ConsoleLogger(name="stderr", console=sys.stderr)
        sys.stderr = stderr_console_logger

        # -- Training context ----------------------------------------------------

        if cloud_mode:
            # Training is running in the cloud
            context = backend.get_cloud_training_context()
        else:
            # Training is running locally
            context = backend.get_local_training_context(force_download)

        # -- Train Parameters ----------------------------------------------------

        params = {}
        params.update(context.get_parameters(model))
        params.update(dict(param.split(":") for param in parameters))
        params["seed"] = seed if seed is not None else model.generate_seed()

        # -- Tags ----------------------------------------------------------------

        tags = {PACKAGE_NAME: str(PACKAGE_VERSION)}
        tags.update(params)
        tags.update(context.get_tags(model))
        tags.update(model.extra_tags())

        # -- Training ------------------------------------------------------------

        print("Training model...")

        try:
            trained_model = context.train_model(model=model, tags=tags, **params)
        except NotImplementedError:
            print("Training script needs an implementation!")
            print(
                "The train() method of your model raised a NotImplementedError "
                "which indicates that you have not yet implemented it."
            )
            print("Stacktrace follows:")
            print("-" * 79)
            raise

        # -- Verify returned object ----------------------------------------------

        print("-" * 79)
        print("Training complete")
        print("Verifying trained model...")

        # Must be of type TrainedModel
        if not isinstance(trained_model, TrainedModel):
            print("-" * 79)
            print(
                "The object returned by your train()-method must be of type "
                "TrainedModel (or inherited classes). "
                f"You gave me something of type {type(trained_model)} instead."
            )
            raise click.Abort()

        # Verify object properties
        try:
            trained_model.verify()
        except trained_model.Invalid as ex:
            print("-" * 79)
            print(f"{trained_model.__class__.__name__} does not validate: {ex}")
            raise click.Abort()

        # -- Dump output to disk -------------------------------------------------

        print(f"Dumping trained model to: {model.trained_model_path}")

        trained_model.params.update(params)
        model.dump(model.trained_model_path, trained_model)

        meta_data = {"module_name": model.module_name, "run_id": context.az_run.id}

        if context.az_run:
            workspace = context.az_run.experiment.workspace
            meta_data["subscription_id"] = workspace.subscription_id
            meta_data["resource_group"] = workspace.resource_group
            meta_data["workspace_name"] = workspace.name
            meta_data["run_id"] = context.az_run.id
            meta_data["portal_url"] = context.get_portal_url()

        context.save_meta_data(
            meta_data, os.path.join(model.artifact_path, model._META_FILE_NAME)
        )

        # -- Upload output files -------------------------------------------------

        print("Uploading output files...")

        with click_spinner.spinner():
            context.save_artifacts(model)

    except Exception:
        raise
    finally:
        if context.az_run:
            # -- Print portal link ---------------------------------------------------
            print(f"Portal link: {context.get_portal_url()}")

            print("Uploading log files...")
            with click_spinner.spinner():
                context.save_logs(stdout_console_logger, stderr_console_logger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage cloud infrastructure for machine learning projects.
"""
import click

from .init import init_infrastructure


@click.group()
def infrastructure_group():
    """Manage cloud infrastructure for machine learning projects."""
    pass


infrastructure_group.add_command(init_infrastructure, "init")



import pytest

from energinetml.core.backend import AbstractBackend


@pytest.fixture
def backend():
    yield AbstractBackend()


class TestAbstractBackend:
    def test__get_available_subscriptions(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_available_subscriptions()

    def test__get_available_resource_groups(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_available_resource_groups("subscription_id")

    def test__get_available_workspaces(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_available_workspaces("subscription_id", "resource_group")

    def test__get_available_workspace_names(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_available_workspace_names("subscription_id", "resource_group")

    def test__get_workspace(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            project_meta = {
                "subscription_id": "subscription_id",
                "resource_group": "resource_group",
                "workspace_name": "name",
            }

            backend.get_workspace(project_meta)

    def test__get_compute_clusters(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_compute_clusters("workspace")

    def test__get_available_vm_sizes(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_available_vm_sizes("workspace")

    def test__create_compute_cluster(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.create_compute_cluster(
                "workspace",
                "name",
                "vm_size",
                "min_nodes",
                "max_nodes",
                "vnet_resource_group_name",
                "vnet_name",
                "subnet_name",
            )

    def test__get_local_training_context(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_local_training_context("force_download")

    def test__get_cloud_training_context(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.get_cloud_training_context()

    def test__submit_model(self, backend):
        """
        :param AbstractBackend backend:
        """
        with pytest.raises(NotImplementedError):
            backend.submit_model("model", "params")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from __future__ import annotations

import os
from typing import Callable, Iterable, Union

import click

from energinetml.backend import default_backend as backend
from energinetml.core.backend import BackendException

# -- Helper classes ----------------------------------------------------------


class ArbitraryChoice(click.Choice):
    """
    An implementation of Click.Choice (for prompting for values) that allows
    the user to enter values which is outside the pre-defined range of choices.

    The pre-defined range of choices can be considered suggestions instead of
    absolute options.
    """

    def get_missing_message(self, param: click.Parameter) -> str:
        """[summary]

        Args:
            param (click.Parameter): [description]

        Returns:
            str: [description]
        """
        return f"{super().get_missing_message(param)} (or enter another value)"

    def convert(
        self,
        value: Union[str, int, float, bool],
        param: click.Parameter,
        ctx: click.Context,
    ) -> ArbitraryChoice:
        """[summary]

        Args:
            value (Union[str, int, float, bool]): [description]
            param (click.Parameter): [description]
            ctx (click.Context): [description]

        Returns:
            ArbitraryChoice: [description]
        """
        try:
            return super().convert(value, param, ctx)
        except click.BadParameter:
            return value


# -- Command parameter parsing -----------------------------------------------


def parse_input_path(project_files: Iterable[str]) -> Callable:
    """[summary]

    Args:
        project_files (Iterable[str]): [description]

    Returns:
        Callable: [description]
    """

    def _parse_input_path(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Raises:
            click.Abort: [description]

        Returns:
            str: [description]
        """
        if value is None:
            value = os.path.abspath(
                click.prompt(
                    text="Enter project location",
                    default=os.path.abspath("."),
                    type=click.Path(dir_okay=True, resolve_path=True),
                )
            )

        # Path points to a file?
        if os.path.isfile(value):
            click.echo("Failed to init project.")
            click.echo(
                "The path you provided me with points to a file, and not a "
                "folder. I need a folder to put the project files in. "
                "Check your -p/--path parameter."
            )
            click.echo("You provided me with: %s" % value)
            raise click.Abort()

        # Confirm overwrite files if they exists
        for filename in project_files:
            if os.path.isfile(os.path.join(value, filename)):
                click.echo("File already exists: %s" % os.path.join(value, filename))
                if not click.confirm("Really override existing %s?" % filename):
                    raise click.Abort()

        return value

    return _parse_input_path


def parse_input_project_name() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_input_project_name(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        if value is None:
            default = (
                os.path.split(ctx.params["path"])[1] if ctx.params.get("path") else None
            )

            value = click.prompt(
                text="Please enter a project name", default=default, type=str
            )

        return value

    return _parse_input_project_name


def parse_input_subscription_id() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_input_subscription_id(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        subscriptions = backend.get_available_subscriptions()
        subscriptions_mapped = {
            s.display_name: s.subscription_id for s in subscriptions
        }

        if value not in subscriptions_mapped:
            if value is not None:
                click.echo(f"Azure Subscription '{value}' not found")

            value = click.prompt(
                text="Please enter Azure Subscription",
                type=click.Choice(subscriptions_mapped.keys()),
            )

        return subscriptions_mapped[value]

    return _parse_input_subscription_id


def parse_input_resource_group() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_input_resource_group(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        if "subscription_id" in ctx.params:
            subscription_id = ctx.params["subscription_id"]
            resource_groups = backend.get_available_resource_groups(subscription_id)
            resource_group_names = [g.name for g in resource_groups]
        else:
            resource_group_names = []

        if value is None or (
            resource_group_names and value not in resource_group_names
        ):

            if value is not None:
                click.echo(f"Azure Resource Group '{value}' not found")

            value = click.prompt(
                text="Please enter Azure Resource Group",
                default="",
                type=(
                    click.Choice(resource_group_names) if resource_group_names else str
                ),
            )

        return value

    return _parse_input_resource_group


def parse_input_workspace_name() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def parse_input_workspace_name(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Raises:
            RuntimeError: [description]
            click.Abort: [description]

        Returns:
            str: [description]
        """
        if "subscription_id" not in ctx.params:
            raise RuntimeError('Requires a "subscription_id" parameter')
        if "resource_group" not in ctx.params:
            raise RuntimeError('Requires a "resource_group" parameter')

        if value is None:
            existing_workspaces = backend.get_available_workspace_names(
                subscription_id=ctx.params["subscription_id"],
                resource_group=ctx.params["resource_group"],
            )

            value = click.prompt(
                text="Please enter AzureML Workspace name",
                default=existing_workspaces[0] if existing_workspaces else None,
                type=ArbitraryChoice(existing_workspaces)
                if existing_workspaces
                else str,  # noqa: E501
            )

        project_meta = {
            "subscription_id": ctx.params["subscription_id"],
            "resource_group": ctx.params["resource_group"],
            "workspace_name": value,
        }
        try:
            backend.get_workspace(project_meta)
        except BackendException:
            click.echo(
                'Workspace "{}" not found in resource group: {}'.format(
                    value, ctx.params["resource_group"]
                )
            )
            raise click.Abort()

        return value

    return parse_input_workspace_name


def parse_input_service_connection() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_input_service_connection(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        if value is None:
            value = click.prompt(
                text="Please enter DevOps Service Connection name", default="", type=str
            )
        return value

    return _parse_input_service_connection


def parse_input_webapp_kind() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_input_webapp_kind(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        if value is None:
            click.echo("-" * 79)
            click.echo("Select which kind of webserver you will use.")
            click.echo("This option can be changed later.")
            click.echo("")
            click.echo("For apps written in flask, Django, or Dash select WSGI.")
            click.echo("For apps written in fastapi select ASGI.")
            click.echo("")
            value = click.prompt(
                text="Please enter webserver kind",
                default=None,
                type=click.Choice(["ASGI", "WSGI"]),
            )
        return value

    return _parse_input_webapp_kind


def parse_pipeline_repo_version() -> Callable:
    """[summary]

    Returns:
        Callable: [description]
    """

    def _parse_pipeline_repo_version(
        ctx: click.Context, param: click.Parameter, value: str
    ) -> str:
        """[summary]

        Args:
            ctx (click.Context): [description]
            param (click.Parameter): [description]
            value (str): [description]

        Returns:
            str: [description]
        """
        if value is None:
            value = click.prompt(
                text="Please enter a version of the pipeline repository",
                default="v0.7.0",
                type=str,
            )
        return value

    return _parse_pipeline_repo_version



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import logging
import time
import traceback
from contextvars import ContextVar
from typing import TYPE_CHECKING, Callable, List, Optional  # noqa TYP001
from uuid import uuid4

import fastapi
import uvicorn
from opentelemetry import trace
from pydantic import BaseModel
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import RedirectResponse

from energinetml.core.predicting import PredictionController, PredictResponse
from energinetml.settings import APPINSIGHTS_INSTRUMENTATIONKEY, PACKAGE_REQUIREMENT

if TYPE_CHECKING:
    from energinetml.core.model import Model, TrainedModel

logger = logging.getLogger(__name__)
tracer = trace.get_tracer(__name__)
correlation_id: ContextVar[str] = ContextVar("correlation_id", default="")


class ModelMetaData(BaseModel):
    """A meta data model container which holds information about the deployed model."""

    model_version: str
    model_name: str
    model_experiment_id: str
    datasets: List[str]
    has_validator: bool
    subscription_id: Optional[str] = None
    resource_group: Optional[str] = None
    workspace_name: Optional[str] = None
    run_id: Optional[str] = None
    portal_url: Optional[str] = None


def create_app(
    model: "Model",
    trained_model: "TrainedModel",
    model_meta_data: ModelMetaData,
    model_version: str = None,
) -> fastapi.FastAPI:
    """A function which defines the FastAPI structure.

    Args:
        model (Model): A Model object which is used and enriched during model training
        with our SDK.
        trained_model (TrainedModel): A TrainedModel object which is generated during
        model training with our SDK.
        model_meta_data (ModelMetaData, optional): A metadata container which is
        exposed on /metadata and can be used for model tracing..
        model_version (str, optional): A model identifer which is genereated during
        registration of an model. Defaults to None.

    Returns:
        fastapi.FastAPI: The structure of the webserver.
    """
    controller = PredictionController(
        model=model, trained_model=trained_model, model_version=model_version
    )

    async def opentelemetry_middleware(
        request: fastapi.Request, call_next: Callable
    ) -> fastapi.Response:
        """
        FastAPI middleware to record HTTP requests.

        Can not access request body in middleware (for logging):
        Issue description: https://github.com/tiangolo/fastapi/issues/394
        """
        start = time.perf_counter()
        correlation_id.set(str(uuid4()))

        # TODO: Why imports here?
        from opentelemetry.trace import SpanKind, Status
        from opentelemetry.trace.status import StatusCode

        start_span = tracer.start_span(name="request", kind=SpanKind.SERVER)

        with start_span as span:
            span.set_attribute("http.url", str(request.url))
            span.set_attribute("http_url", str(request.url))
            span.set_attribute("model_name", model.name)
            span.set_attribute("correlation_id", correlation_id.get())
            if model_version is not None:
                span.set_attribute("model_version", model_version)

            try:
                response = await call_next(request)
            except Exception as e:
                logger.exception("Prediction failed")
                span.record_exception(e)
                span.set_status(Status(status_code=StatusCode.ERROR))
                span.set_attribute("http.status_code", 500)
                span.set_attribute("http_status_code", 500)
                span.set_attribute("error.name", e.__class__.__name__)
                span.set_attribute("error.message", str(e))
                span.set_attribute("error.stacktrace", traceback.format_exc())
                return fastapi.Response(status_code=500)
            else:
                span.set_status(Status(status_code=StatusCode.OK))
                span.set_attribute("http.status_code", response.status_code)
                span.set_attribute("http_status_code", response.status_code)
                end = time.perf_counter()
                span.set_attribute("duration_endpoint", str(end - start))
                return response

    async def root_http_endpoint() -> RedirectResponse:
        """
        / should redirect to /docs
        """
        return RedirectResponse(url="/docs")

    async def predict_http_endpoint(
        request: controller.request_model, response: fastapi.Response
    ) -> PredictResponse:
        """
        Model prediction HTTP endpoint.
        """
        response.headers["X-sdk-version"] = str(PACKAGE_REQUIREMENT)

        return controller.predict(request, correlation_id=correlation_id.get())

    async def health_http_endpoint(response: fastapi.Response) -> fastapi.Response:
        """
        Health endpoint, should return status 200 with no specific body.
        """
        response.status_code = 200
        return response

    async def metadata_http_endpoint(response: fastapi.Response) -> ModelMetaData:
        """
        Metadata endpoint, should return status 200 with a ModelMetaData body.
        """
        return model_meta_data

    # -- Setup app -----------------------------------------------------------

    if model_meta_data.portal_url:
        description = (
            "This model was trained via AzureML. The run can be found "
            f"<a href='{model_meta_data.portal_url}' target='_blank'>here</a>."
            f"\n\nModel version: {model_version}"
        )
    else:
        description = f"Model version: {model_version if model_version else None}"

    app = fastapi.FastAPI(
        title=model.name,
        description=description,
    )

    if APPINSIGHTS_INSTRUMENTATIONKEY:
        app.add_middleware(
            middleware_class=BaseHTTPMiddleware, dispatch=opentelemetry_middleware
        )

    app.router.add_api_route(path="/", methods=["GET"], endpoint=root_http_endpoint)

    app.router.add_api_route(
        path="/predict",
        methods=["POST"],
        endpoint=predict_http_endpoint,
        response_model=controller.response_model,
        tags=["model"],
        summary="Predict using the model",
    )

    app.router.add_api_route(
        path="/health",
        methods=["GET"],
        endpoint=health_http_endpoint,
        tags=["health"],
        summary="Health endpoint",
        description="Health endpoint, returns status 200",
    )

    app.router.add_api_route(
        path="/metadata",
        methods=["GET"],
        endpoint=metadata_http_endpoint,
        tags=["Metadata"],
        summary="Metadata endpoint",
        description="Metadata endpoint, returns status 200 with a ModelMetaData body",
    )

    return app


def run_predict_api(
    model: "Model",
    trained_model: "TrainedModel",
    meta_data,
    host: str,
    port: int,
    model_version: str = None,
) -> None:
    """Wrapper function used to enrich the defined endpoints.

    Args:
        model (Model): A Model object which is used and enriched during model training
        with our SDK.
        trained_model (TrainedModel): A TrainedModel object which is generated during
        model training with our SDK.
        host (str): Name of the host. This value is "127.0.0.1" when running local.
        port (int): Port number where the webserver is exposing defined endpoints to.
        model_version (str, optional): A model identifer which is genereated during
        registration of an model. Defaults to None.
    """

    model_meta_data = ModelMetaData(
        model_version=model_version,
        model_name=model.name,
        model_experiment_id=model.experiment,
        has_validator=True if trained_model.validator else False,
        datasets=model.datasets,
        subscription_id=meta_data["subscription_id"],
        resource_group=meta_data["resource_group"],
        workspace_name=meta_data["workspace_name"],
        run_id=meta_data["run_id"],
        portal_url=meta_data["portal_url"],
    )

    app = create_app(
        model=model,
        trained_model=trained_model,
        model_version=model_version,
        model_meta_data=model_meta_data,
    )

    uvicorn.run(app=app, host=host, port=port)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage machine learning projects.
"""
import click

from .init import init_project


@click.group()
def project_group() -> None:
    """Manage machine learning projects."""
    pass


project_group.add_command(init_project, "init")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

import os
import subprocess
import sys
import tempfile
import typing

from jinja2 import Template

from energinetml.settings import (
    DEFAULT_ENCODING,
    DEVOPS_FOLDER_NAME,
    GITIGNORE_NAME,
    TEMPLATES_GIT_URL,
    TEMPLATES_IP_WHITELIST,
    TEMPLATES_SUBNET_WHITELIST,
)


class TemplateResolver:
    """[summary]"""

    class TemplateResolverError(Exception):
        """[summary]"""

        pass

    def get_default_env(self) -> typing.Dict[str, typing.Any]:
        """[summary]

        Returns:
            typing.Dict[str, typing.Any]: [description]
        """
        return {}

    def clone_and_render(
        self,
        project_root_path: str,
        files: typing.Iterable[typing.Tuple[str, str]],
        env: typing.Dict[str, typing.Any],
    ):
        """Clones Git repository and renders templates using the provided
        environment variables.

        Args:
            project_root_path (str): [description]
            files (typing.Iterable[typing.Tuple[str, str]]): This argument is an
            iterable of (src, dst) where 'src' is file path relative to Git repository
            root, and 'dst' is file path relative.
        to project root.
            env (typing.Dict[str, typing.Any]): [description]
        """
        actual_env = self.get_default_env()
        actual_env.update(env)

        with tempfile.TemporaryDirectory() as clone_path:
            self.clone(clone_path)

            for src, dst in files:
                self.render(
                    src=os.path.join(clone_path, src),
                    dst=os.path.join(project_root_path, dst),
                    env=actual_env,
                )

    def clone(self, clone_path: str) -> None:
        """[summary]

        Args:
            clone_path (str): [description]

        Raises:
            self.TemplateResolverError: [description]
        """
        try:
            subprocess.check_call(
                args=["git", "clone", TEMPLATES_GIT_URL, clone_path],
                stdout=sys.stdout,
                stderr=subprocess.STDOUT,
                shell=False,
            )
        except subprocess.CalledProcessError:
            raise self.TemplateResolverError(
                f"Failed to clone Git repo: {TEMPLATES_GIT_URL}"
            )

    def render(self, src: str, dst: str, env: typing.Dict[str, str]) -> None:
        """Renders Jinja2 template file at 'src' and writes it to 'dst' using the
        provided environment variables. Creates directories if necessary.

        Args:
            src (str): [description]
            dst (str): [description]
            env (typing.Dict[str, str]): [description]
        """
        dst_folder = os.path.split(dst)[0]

        with open(src, encoding=DEFAULT_ENCODING) as f:
            template = Template(f.read())
            rendered = template.render(**env)

        if not os.path.isdir(dst_folder):
            os.makedirs(dst_folder)

        with open(dst, "w", encoding=DEFAULT_ENCODING) as f:
            f.write(rendered)


class DataSciencePlatformTemplates(TemplateResolver):
    """
    Templates specific for Data Science Platform.
    """

    def get_default_env(self) -> typing.Dict[str, typing.Any]:
        """[summary]

        Returns:
            typing.Dict[str, typing.Any]: [description]
        """
        return {
            "subnetWhitelist": TEMPLATES_SUBNET_WHITELIST,
            "ipWhitelist": TEMPLATES_IP_WHITELIST,
        }

    def resolve(
        self,
        project_root_path: str,
        project_name: str,
        service_connection: str,
        resource_group: str,
        pipeline_repo_version: str,
    ):
        """[summary]

        Args:
            project_root_path (str): [description]
            project_name (str): [description]
            service_connection (str): DevOps Service Connection for deploying to
            resource group.
            resource_group (str): Azure Resource Group to deploy webapp to.
        """
        files = (
            (GITIGNORE_NAME, GITIGNORE_NAME),
            (
                os.path.join(DEVOPS_FOLDER_NAME, "infrastructure.yml"),
                os.path.join(DEVOPS_FOLDER_NAME, "infrastructure.yml"),
            ),
            (
                os.path.join("terraform", "datascienceplatform", "dev", "main.tf"),
                os.path.join("terraform", "dev", "datascienceplatform.tf"),
            ),
        )

        env = {
            "serviceConnection": service_connection,
            "resourceGroup": resource_group,
            "projectName": project_name,
            "pipelineRepoVersion": pipeline_repo_version,
        }

        self.clone_and_render(project_root_path=project_root_path, files=files, env=env)


class WebAppTemplateResolver(TemplateResolver):
    """
    Templates specific for Web Apps.
    """

    def resolve_web_app(
        self, project_root_path: str, kind: str, env: typing.Dict[str, typing.Any]
    ):
        """[summary]

        Args:
            project_root_path (str): [description]
            kind (str): [description]
            env (typing.Dict[str, typing.Any]): [description]
        """
        files = (
            (GITIGNORE_NAME, GITIGNORE_NAME),
            (
                os.path.join(DEVOPS_FOLDER_NAME, "infrastructure.yml"),
                os.path.join(DEVOPS_FOLDER_NAME, "infrastructure.yml"),
            ),
            (
                os.path.join(DEVOPS_FOLDER_NAME, "deploy-webapp.yml"),
                os.path.join(DEVOPS_FOLDER_NAME, "deploy.yml"),
            ),
            (
                os.path.join("webapp", "terraform", "dev", "main.tf"),
                os.path.join("terraform", "dev", "webapp.tf"),
            ),
            (os.path.join("webapp", kind, "Dockerfile"), os.path.join("Dockerfile")),
            (
                os.path.join("webapp", kind, "src", "app.py"),
                os.path.join("src", "app.py"),
            ),
            (
                os.path.join("webapp", kind, "src", "__init__.py"),
                os.path.join("src", "__init__.py"),
            ),
            (
                os.path.join("webapp", kind, "src", "requirements.txt"),
                os.path.join("src", "requirements.txt"),
            ),
        )

        self.clone_and_render(project_root_path=project_root_path, files=files, env=env)

    def resolve(
        self,
        project_root_path: str,
        project_name: str,
        service_connection: str,
        resource_group: str,
        pipeline_repo_version: str,
    ):
        """[summary]

        Args:
            project_root_path (str): [description]
            project_name (str): [description]
            service_connection (str): DevOps Service Connection for deploying to
            resource group.
            resource_group (str): Azure Resource Group to deploy webapp to.

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError


class ASGIWebAppTemplates(WebAppTemplateResolver):
    """
    Templates specific for ASGI Web Apps.
    """

    def resolve(
        self,
        project_root_path: str,
        project_name: str,
        service_connection: str,
        resource_group: str,
        pipeline_repo_version: str,
    ):
        """[summary]

        Args:
            project_root_path (str): [description]
            project_name (str): [description]
            service_connection (str): DevOps Service Connection for deploying to
            resource group.
            resource_group (str): Azure Resource Group to deploy webapp to.
        """
        self.resolve_web_app(
            project_root_path=project_root_path,
            kind="ASGI",
            env={
                "serviceConnection": service_connection,
                "resourceGroup": resource_group,
                "projectName": project_name,
                "pipelineRepoVersion": pipeline_repo_version,
            },
        )


class WSGIWebAppTemplates(WebAppTemplateResolver):
    """
    Templates specific for ASGI Web Apps.
    """

    def resolve(
        self,
        project_root_path: str,
        project_name: str,
        service_connection: str,
        resource_group: str,
        pipeline_repo_version: str,
    ):
        """[summary]

        Args:
            project_root_path (str): [description]
            project_name (str): [description]
            service_connection (str): DevOps Service Connection for deploying to
            resource group.
            resource_group (str): Azure Resource Group to deploy webapp to.
        """
        self.resolve_web_app(
            project_root_path=project_root_path,
            kind="WSGI",
            env={
                "serviceConnection": service_connection,
                "resourceGroup": resource_group,
                "projectName": project_name,
                "pipelineRepoVersion": pipeline_repo_version,
            },
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict  # noqa TYP001

if TYPE_CHECKING:
    from energinetml import Model, TrainedModel

from energinetml.core.logger import ConsoleLogger


class TrainingError(Exception):
    pass


class AbstractTrainingContext:
    """[summary]"""

    def train_model(
        self, model: Model, tags: Dict[str, Any], *args: Any, **kwargs: Dict[str, Any]
    ) -> TrainedModel:
        """[summary]

        Args:
            model (Model): [description]
            tags (Dict[str, Any]): [description]

        Returns:
            TrainedModel: [description]
        """
        pass

    def save_artifacts(self, model: Model) -> None:
        """Saves output files from a training run if necessary.

        Args:
            model (Model): [description]
        """
        pass

    def get_parameters(self, model: Model) -> Dict[str, str]:
        """Returns parameters for a training.

        Args:
            model (Model): [description]

        Returns:
            Dict[str, str]: [description]
        """
        return {}

    def get_tags(self, model: Model) -> Dict[str, str]:
        """Returns tags for a training.

        Args:
            model (Model): [description]

        Returns:
            Dict[str, str]: [description]
        """
        return {}

    def save_logs(self, clog: ConsoleLogger) -> None:
        """This function takes the log file generated from clog and
        pushes the log into the azure ml expiremnt tab called output.

        Args:
            clog (ConsoleLogger): This argument is an object of our logger function.

        """
        pass

    def save_meta_data(self, meta_data: Dict[str, str], meta_file_path: str):
        """Create a meta json-file with the content from a meta_data dictionary
        and save it to the path specified.

        Args:
            meta_data (Dict[str, str])): A dictionary of meta_data to persist in the
            json-file
            meta_file_path (str): The local path where the json-file is persisted.

        """
        pass


def requires_parameter(name, typ):
    """[summary]

    Args:
        name ([type]): [description]
        typ ([type]): [description]
    """

    def requires_parameter_decorator(func):
        """[summary]

        Args:
            func ([type]): [description]
        """

        def requires_parameter_inner(*args: Any, **kwargs: Dict[str, Any]):
            """[summary]

            Raises:
                TrainingError: [description]
                TrainingError: [description]

            Returns:
                [type]: [description]
            """
            if name not in kwargs:
                raise TrainingError(f'Missing parameter "{name}"')
            try:
                kwargs[name] = typ(kwargs.get(name))
            except ValueError:
                raise TrainingError(
                    (
                        f'Parameter "{name}" could not be cast '
                        f"to type { typ.__name__}: { kwargs.get(name)}"
                    )
                )
            return func(*args, **kwargs)

        return requires_parameter_inner

    return requires_parameter_decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""

from __future__ import annotations

import os
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Any, Dict

from energinetml.core.templates import WebAppTemplateResolver
from energinetml.core.webapp import WebAppEngine
from energinetml.settings import (
    DEFAULT_ENCODING,
    DEFAULT_LOCATION,
    PACKAGE_NAME,
    PACKAGE_VERSION,
)

from .configurable import Configurable
from .requirements import RequirementList
from .templates import ASGIWebAppTemplates, WSGIWebAppTemplates
from .webapp import AsgiWebApp, WsgiWebApp


@dataclass
class Project(Configurable):
    """[summary]"""

    _CONFIG_FILE_NAME = "project.json"
    _REQUIREMENTS_FILE_NAME = "requirements.txt"

    name: str

    @classmethod
    def create(cls, *args: Any, **kwargs: Dict[str, Any]) -> Project:
        """[summary]

        Returns:
            [type]: [description]
        """
        project = super().create(*args, **kwargs)

        # Create requirements.txt file
        with open(project.requirements_file_path, "w", encoding=DEFAULT_ENCODING) as f:
            f.write(f"{PACKAGE_NAME}=={PACKAGE_VERSION}\n")

        return project

    @property
    def requirements_file_path(self) -> str:
        """Absolute path to requirements.txt file.

        Returns:
            str: [description]
        """
        return self.get_file_path(self._REQUIREMENTS_FILE_NAME)

    @cached_property
    def requirements(self) -> RequirementList:
        """Returns a list of project requirements from requirements.txt.

        Returns:
            RequirementList: [description]
        """
        if os.path.isfile(self.requirements_file_path):
            return RequirementList.from_file(self.requirements_file_path)
        else:
            return RequirementList()


# -- Machine Learning --------------------------------------------------------


@dataclass
class MachineLearningProject(Project):
    """[summary]"""

    subscription_id: str
    resource_group: str
    workspace_name: str
    vnet_name: str
    subnet_name: str
    location: str = field(default=DEFAULT_LOCATION)

    @property
    def vnet_resourcegroup_name(self) -> str:
        """The resource group where the VNET is located, typically
        the same as the workspace.

        Returns:
            str: [description]
        """
        return self.resource_group

    def default_model_path(self, model_name: str) -> str:
        """Returns default absolute path to folder where new models
        should be created at.

        Args:
            model_name (str): [description]

        Returns:
            str: [description]
        """
        return self.get_file_path(model_name)


# -- Web Apps ----------------------------------------------------------------


class WebAppProjectKind(str, Enum):
    """[summary]"""

    ASGI = "ASGI"
    WSGI = "WSGI"


@dataclass
class WebAppProject(Project):
    """[summary]"""

    kind: WebAppProjectKind

    RESOLVERS = {
        WebAppProjectKind.ASGI: ASGIWebAppTemplates(),
        WebAppProjectKind.WSGI: WSGIWebAppTemplates(),
    }

    ENGINES = {
        WebAppProjectKind.ASGI: AsgiWebApp(),
        WebAppProjectKind.WSGI: WsgiWebApp(),
    }

    def get_template_resolver(self) -> WebAppTemplateResolver:
        """[summary]

        Raises:
            RuntimeError: [description]

        Returns:
            WebAppTemplateResolver: [description]
        """
        if self.kind in self.RESOLVERS:
            return self.RESOLVERS[self.kind]
        raise RuntimeError

    def get_engine(self) -> WebAppEngine:
        """[summary]

        Raises:
            RuntimeError: [description]

        Returns:
            WebAppEngine: [description]
        """
        if self.kind in self.ENGINES:
            return self.ENGINES[self.kind]
        raise RuntimeError

    @property
    def dockerfile_path(self) -> str:
        """[summary]

        Returns:
            str: [description]
        """
        return os.path.join(self.path, "Dockerfile")



import os
import tempfile
from unittest.mock import Mock, patch

import click
import pytest
from click.testing import CliRunner

from energinetml.cli.utils.decorators import discover_model, discover_project
from energinetml.core.model import (
    Model,
    ModelImportError,
    ModelNotClassError,
    ModelNotInheritModel,
)
from energinetml.core.project import MachineLearningProject

PROJECT_NAME = "NAME"
PROJECT_SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
PROJECT_RESOURCE_GROUP = "RESOURCE-GROUP"
PROJECT_WORKSPACE_NAME = "WORKSPACE-NAME"

MODEL_NAME = "NAME"
MODEL_EXPERIMENT = "EXPERIMENT"
MODEL_COMPUTE_TARGET = "COMPUTE-TARGET"
MODEL_VM_SIZE = "VM-SIZE"
MODEL_DATASETS = ["iris", "hades:2"]
MODEL_FEATURES = ["feature1", "feature2"]
MODEL_PARAMETERS = {"param1": "value1", "param2": "value2"}


# -- discover_project() Tests ------------------------------------------------


def test__discover_project__project_exists__should_exit_with_status_ok():
    runner = CliRunner()

    @click.command()
    @discover_project(Mock())
    def discover_project_testable(project):
        pass

    # Act
    result = runner.invoke(discover_project_testable, ["--path", "mock-path"])

    # Assert
    assert result.exit_code == 0, str(result.exception)


def test__discover_project__project_does_not_exist__should_exit_with_error():
    with tempfile.TemporaryDirectory() as path:

        runner = CliRunner()
        project_cls_mock = Mock()
        project_cls_mock.from_directory.side_effect = (
            MachineLearningProject.ConfigNotFound
        )

        @click.command()
        @discover_project(project_cls_mock)
        def discover_project_testable(project):
            pass

        # Act
        result = runner.invoke(discover_project_testable, ["--path", path])

        # Assert
        assert result.exit_code == 1, str(result.exception)

        project_cls_mock.from_directory.assert_called_once_with(path)


# -- discover_model() Tests --------------------------------------------------


@click.command()
@discover_model()
def discover_model_testable(model):
    """
    Empty function to allow invoking the decorator.
    """
    pass


@patch("energinetml.cli.utils.decorators.os.path.isdir")
def test__discover_model__model_file_does_not_exists__should_exit_with_error(
    isdir_mock,
):  # noqa: E501
    """
    :param Mock isdir_mock:
    """
    with tempfile.TemporaryDirectory() as path:
        isdir_mock.return_value = False

        runner = CliRunner()
        # Act
        result = runner.invoke(discover_model_testable, ["--path", path])

        # Assert
        assert result.exit_code == 1
        isdir_mock.assert_called_once_with(os.path.normpath(path))


@patch("energinetml.cli.utils.decorators.os.path.isdir")
@patch("energinetml.cli.utils.decorators.import_model_class")
@patch("energinetml.core.configurable.Configurable.from_directory")
@pytest.mark.parametrize(
    "exception", (ModelImportError, ModelNotClassError, ModelNotInheritModel)
)
def test__discover_model__import_model_class_raises_exception__should_exit_with_error(
    from_directory_mock, import_model_class_mock, isdir_mock, exception
):
    """
    :param Mock import_model_class_mock:
    :param Mock isdir_mock:
    :param Exception exception:
    """
    parent_levels = 1
    with tempfile.TemporaryDirectory() as path:
        isdir_mock.return_value = True
        from_directory_mock.return_value = Mock(parent_levels=parent_levels)
        import_model_class_mock.side_effect = exception

        runner = CliRunner()

        # Act
        result = runner.invoke(discover_model_testable, ["--path", path])

        # Assert
        assert result.exit_code == 1

        isdir_mock.assert_called_once_with(os.path.normpath(path))
        import_model_class_mock.assert_called_once_with(path, parent_levels)


@patch("energinetml.cli.utils.decorators.os.path.isdir")
@patch("energinetml.cli.utils.decorators.import_model_class")
@patch("energinetml.core.configurable.Configurable.from_directory")
def test__discover_model__model_not_found_in_directory__should_exit_with_error(
    from_directory_mock, import_model_class_mock, isdir_mock
):
    """
    :param Mock import_model_class_mock:
    :param Mock isdir_mock:
    """
    parent_levels = 1
    with tempfile.TemporaryDirectory() as path:
        from_directory_mock.return_value = Mock(parent_levels=parent_levels)
        model_class = Mock()
        model_class.from_directory.side_effect = Model.ConfigNotFound
        isdir_mock.return_value = True
        import_model_class_mock.return_value = model_class

        runner = CliRunner()

        # Act
        result = runner.invoke(discover_model_testable, ["--path", path])

        # Assert
        assert result.exit_code == 1

        isdir_mock.assert_called_once_with(os.path.normpath(path))

        import_model_class_mock.assert_called_once_with(path, parent_levels)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
from .decorators import discover_model, discover_project
from .projects import (
    parse_input_path,
    parse_input_project_name,
    parse_input_resource_group,
    parse_input_service_connection,
    parse_input_subscription_id,
    parse_input_webapp_kind,
    parse_input_workspace_name,
    parse_pipeline_repo_version,
)



import json
import os

from energinetml.azure.training import AzureTrainingContext
from energinetml.core.model import Model
from tests.constants import RUN_ID


def test__should_write_metafile_correct(model):

    # Arrange
    model_meta_data = {"module_name": model.module_name, "run_id": RUN_ID}
    context = AzureTrainingContext()
    file_path = os.path.join(model.path, Model._META_FILE_NAME)

    # Act
    context.save_meta_data(model_meta_data, file_path)

    # Read file from disk
    with open(file_path) as file:
        meta_dict = json.load(file)

    # Assert
    assert meta_dict["module_name"] == model.module_name
    assert meta_dict["run_id"] == RUN_ID



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import os

from packaging import version
from pkg_resources import Requirement

# -- Directories/paths -------------------------------------------------------

__current_file = os.path.abspath(__file__)
__current_folder = os.path.split(__current_file)[0]

SOURCE_DIR = os.path.abspath(__current_folder)
STATIC_DIR = os.path.join(SOURCE_DIR, "static")
EMPTY_MODEL_TEMPLATE_DIR = os.path.join(STATIC_DIR, "model_template")
DOCKERFILE_PATH_ML_MODEL = os.path.join(STATIC_DIR, "Dockerfile")
GITIGNORE_PATH = os.path.join(STATIC_DIR, "gitignore.txt")
WEB_APP_PIPELINES_TEMPLATES_PATH = os.path.join(STATIC_DIR, "webapppipelines")
DEVOPS_FOLDER_NAME = ".azuredevops"
GITIGNORE_NAME = ".gitignore"
DEFAULT_ENCODING = "utf-8"


def __read_meta(filename: str) -> str:
    with open(
        os.path.join(__current_folder, "meta", filename), encoding=DEFAULT_ENCODING
    ) as handle:
        return handle.read().strip()


# -- Local -------------------------------------------------------------------

DEFAULT_RELATIVE_ARTIFACT_PATH = "outputs"
DEFAULT_LOG_PATH = "logs"
DEFAULT_LOG_FILENAME = "log.txt"
DEFAULT_LOG_ENCODING = "utf-8"

# -- Cloud --------------------------------------------------------------------

DEFAULT_LOCATION = "westeurope"
DEFAULT_VM_CPU = "Standard_D1_v2"
DEFAULT_VM_GPU = "Standard_NV6"

CLUSTER_IDLE_SECONDS_BEFORE_SCALEDOWN = 60 * 60 * 2  # 2 hours


# -- Package details ---------------------------------------------------------

# TODO Rename "PACKAGE" to "SDK" (here and elsewhere in general)

PYTHON_VERSION = __read_meta("PYTHON_VERSION")
PACKAGE_NAME = __read_meta("PACKAGE_NAME")
COMMAND_NAME = __read_meta("COMMAND_NAME")
PACKAGE_VERSION = version.parse(__read_meta("PACKAGE_VERSION"))
PACKAGE_REQUIREMENT = Requirement.parse(f"{PACKAGE_NAME}=={PACKAGE_VERSION}")


# -- Misc --------------------------------------------------------------------

APPINSIGHTS_INSTRUMENTATIONKEY = os.environ.get("APPINSIGHTS_INSTRUMENTATIONKEY")

# Git repository containing template files
TEMPLATES_GIT_URL = "https://github.com/AnalyticsOps/templates.git"
TEMPLATES_SUBNET_WHITELIST = "/subscriptions/d252b8da-1db8-44f7-bd6e-8010cc01382b/resourceGroups/rg-AnalyticsOpsAgents-U/providers/Microsoft.Network/virtualNetworks/vnet-devops-agent-001/subnets/agent-subnet"  # noqa: E501
TEMPLATES_IP_WHITELIST = "194.239.2.0/24"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import click

from energinetml.core.docker import build_prediction_api_docker_image


@click.command()
@click.option(
    "--tag",
    "-t",
    required=True,
    help="Name and optionally a tag in the `name:tag` format",
)
@click.option(
    "--model-version",
    "model_version",
    required=True,
    type=str,
    help="Model version (used for logging)",
)
@click.option(
    "--artifact-path",
    "artifact_path",
    required=True,
    type=str,
    help="Path to model artifact",
)
def build(tag: str, model_version: str, artifact_path: str) -> None:
    """Build a Docker image with a HTTP web API for model prediction."""

    build_prediction_api_docker_image(
        artifact_path=artifact_path,
        model_version=model_version,
        tag=tag,
    )



"""
Example setup.py from https://github.com/activescott/python-package-example/blob/master/package-project/src/setup.py.
"""
# Always prefer setuptools over distutils
import os
from distutils.command.sdist import sdist

import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def __read_meta(fn):
    with open(os.path.join(here, "energinetml", "meta", fn)) as f:
        return f.read().strip()


def _write_meta(fn, text):
    with open(os.path.join(here, "energinetml", "meta", fn), "w") as f:
        return f.write(text.strip())


class sdist_hg(sdist):
    """Add git short commit hash to version.
    Based on https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/specification.html
    """

    user_options = sdist.user_options + [
        ("dev", None, "Add a dev marker"),
        ("build=", None, "Build number"),
    ]

    def initialize_options(self):
        sdist.initialize_options(self)
        self.dev = None
        self.build = None

    def run(self):
        if self.build:
            if self.build.startswith("+"):
                prefix = ""
            else:
                prefix = "."
            self.distribution.metadata.version += f"{prefix}{self.build}"
            _write_meta("PACKAGE_VERSION", self.distribution.metadata.version)
            print(self.distribution.metadata.version)
        sdist.run(self)


name = __read_meta("PACKAGE_NAME")
version = __read_meta("PACKAGE_VERSION")
python_requires = __read_meta("PYTHON_VERSION")
command = __read_meta("COMMAND_NAME")


setuptools.setup(
    name=name,
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,
    description="Energinet Machine Learning",
    author="Koncern Digitalisering Advanced Analytics Team",
    author_email="mny@energinet.dk, xjakk@energinet.dk",
    # Choose your license
    license="Apache Software License 2.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        f"Programming Language :: Python :: {python_requires}",
    ],
    # What does your project relate to?
    keywords=[],
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=setuptools.find_packages(),  # Required
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    # TODO SPECIFIC VERSION OF REQUIREMENTS!
    install_requires=[
        "requests",
        "Jinja2==3.0.1",
        "pandas>=1.2.0",
        "packaging==20.9",
        "click==8.0.3",
        "click-spinner==0.1.10",
        "fastapi==0.63.0",
        "uvicorn==0.15.0",
        "pydantic==1.8.1",
        "requirements-parser==0.2.0",
        "azure-cli-core==2.34.1",
        "azure-identity==1.7.0",
        "azureml-core==1.39.0",
        "azureml-dataset-runtime==1.39.0",
        "azure-monitor-opentelemetry-exporter==1.0.0b4",
        "pyopenssl<21.0.0",
        "waitress",
        "flask",
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={"dev": [], "build": [], "test": []},
    cmdclass={"sdist": sdist_hg},
    python_requires=">=3.9.0",
    include_package_data=True,
    package_data={"": ["meta/*", "static/*", "static/model_template/*"]},
    entry_points={"console_scripts": [f"{command}=energinetml.cli:main"]},
)



import tempfile
from unittest.mock import ANY, Mock, PropertyMock, patch

from click.testing import CliRunner

from energinetml.cli.project.init import init_project

NAME = "PROJECT-NAME"
SUBSCRIPTION_NAME = "MY-SUBSCRIPTION"
SUBSCRIPTION_ID = "SUBSCRIPTION-ID"
RESOURCE_GROUP = "RESOURCE-GROUP"
WORKSPACE_NAME = "WORKSPACE-NAME"
VNET = "VNET"
SUBNET = "SUBNET"


# -- init_project() Tests ----------------------------------------------------


@patch("energinetml.cli.utils.projects.backend")
@patch("energinetml.cli.project.init.backend")
@patch("energinetml.cli.project.init.MachineLearningProject.create")
def test__init_project(project_create_mock, backend_mock2, backend_mock):
    """
    :param Mock project_create_mock:
    :param Mock backend_mock2:
    :param Mock backend_mock:
    """

    runner = CliRunner()

    subscription = Mock(subscription_id=SUBSCRIPTION_ID, display_name=SUBSCRIPTION_NAME)

    # Workaround: Can not mock property "name" of Mock object
    resource_group = Mock()
    p = PropertyMock(return_value=RESOURCE_GROUP)
    type(resource_group).name = p

    backend_mock2.get_workspace.return_value = Mock(
        tags={"vnet_name": VNET, "subnet_name": SUBNET}
    )

    backend_mock.get_available_subscriptions.return_value = [subscription]
    backend_mock.get_available_resource_groups.return_value = [resource_group]
    backend_mock.get_available_workspace_names.return_value = [WORKSPACE_NAME]

    with tempfile.TemporaryDirectory() as path:

        # Act
        result = runner.invoke(
            cli=init_project,
            args=[
                "--path",
                path,
                "--name",
                NAME,
                "--subscription",
                SUBSCRIPTION_NAME,
                "--resource-group",
                RESOURCE_GROUP,
                "--workspace",
                WORKSPACE_NAME,
            ],
        )

        # Assert
        assert result.exit_code == 0, str(result.exception)

        project_create_mock.assert_called_once_with(
            path=path,
            name=NAME,
            subscription_id=SUBSCRIPTION_ID,
            resource_group=RESOURCE_GROUP,
            workspace_name=WORKSPACE_NAME,
            location=ANY,
            vnet_name=VNET,
            subnet_name=SUBNET,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""[summary]
"""
import os

import click

from energinetml.cli.cluster import change_cluster, create_cluster
from energinetml.cli.utils import discover_project
from energinetml.core.model import DEFAULT_FILES_EXCLUDE, DEFAULT_FILES_INCLUDE, Model
from energinetml.core.project import MachineLearningProject
from energinetml.settings import COMMAND_NAME

# Constants  TODO Move to Project class?
MODEL_FILES = ("__init__.py", "model.json", "model.py")


# -- Input parsing and validation --------------------------------------------


def _parse_input_model_name(ctx, param, value: str) -> str:
    """[summary]

    Args:
        ctx ([type]): [description]
        param ([type]): [description]
        value (str): [description]

    Returns:
        str: [description]
    """
    if value is None:
        value = click.prompt(text="Please enter a model name", type=str)
    return value


# -- CLI Command -------------------------------------------------------------


@click.command()
@discover_project(MachineLearningProject)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    type=str,
    callback=_parse_input_model_name,
    help="Model name (must be unique in project)",
)
@click.option(
    "--cluster",
    "-c",
    required=False,
    default=None,
    type=click.Choice(["new", "existing", "no"], case_sensitive=False),
    help=(
        "Whether to create a new Compute Cluster for the model, "
        "attach an existing cluster, or no cluster setup."
    ),
)
@click.pass_context
def init_model(ctx, name: str, cluster: bool, project: MachineLearningProject) -> None:
    """Create a new, empty machine learning model."""

    # TODO: Verify name complies to naming conventions
    # TODO: Length of ComputeTarget name must be <= 16

    # -- Target folder -------------------------------------------------------

    path = project.default_model_path(name)

    # Confirm overwrite files if they exists
    for filename in MODEL_FILES:
        if os.path.isfile(os.path.join(path, filename)):
            click.echo("File already exists: %s" % os.path.join(path, filename))
            if not click.confirm("Really override existing %s?" % filename):
                raise click.Abort()

    # Create empty folder is necessary
    if not os.path.isdir(path):
        os.makedirs(path)

    # -- Create model --------------------------------------------------------

    model = Model.create(
        path=path,
        name=name,
        experiment=name,
        compute_target=None,
        vm_size=None,
        files_include=DEFAULT_FILES_INCLUDE,
        files_exclude=DEFAULT_FILES_EXCLUDE,
    )

    click.echo("-" * 79)
    click.echo("Initialized the model at: %s" % path)
    click.echo("")
    click.echo(
        (
            "Next step is to implement your model script! I have created an "
            f"empty model template for you. It is called {Model._SCRIPT_FILE_NAME} "
            "and is located in the model directory. Also, remember to add your model's "
            "dependencies to requirements.txt (located at the root "
            "of your project)."
        )
    )
    click.echo("")
    click.echo(
        f"Use '{COMMAND_NAME} model ...' to interact with your model afterwards."
    )
    click.echo("-" * 79)

    # -- Compute cluster -----------------------------------------------------

    if cluster is None:
        cluster = click.prompt(
            text=(
                "Would you like to setup a new compute cluster for your model, "
                "or use an existing cluster?"
            ),
            type=click.Choice(["new", "existing", "no"], case_sensitive=False),
        )

    if cluster == "new":
        ctx.invoke(create_cluster, model=model)
    elif cluster == "existing":
        ctx.invoke(change_cluster, model=model)
